"""Building module.

This module simulates the thermal behaviour of a building using reference data from the EPISCOPE/TABULA project and
a resistor-capacitor (RC) model from the RC_BuildingSimulator project and the ISO 13790 norm.

The module contains the following classes:
    1. class BuildingState - constructs the building state.
    2. class BuildingControllerState - constructs the building controller state.
    3. class BuildingConfig - json dataclass, configurates the building class.
    4. class BuildingControllerConfig -  json dataclass, configurates the building controller class.
    5. class Building -  main class, involves building properties taken from EPISCOPE/TABULA project database* and
       calculations from RC_BuildingSimulator project**. Inputs of the building class are the heating device, occupancy and weather.
       Outputs are for example temperature, stored energy and solar gains through windows.
    6. class Window - taken from the RC simulator project, calculates for example solar gains through windows.
    7. class BuildingController - calculates real heating demand and how much building is supposed to be heated up.

*EPISCOPE/TABULA project:
     Ths project involves a collection of multiple typologies of residences from 12 European countries, listing among others,
     heat coefficient, area, volumes, light transmissibility, house heat capacity for various residence construction elements.
     These typologies are categorized by year of construction, residence type and degree of refurbishment.
     For information, please access site: https://episcope.eu/building-typology/webtool/

**RC_BuildingSimulator project:
    The functions cited in this module are at some degree based on the RC_BuildingSimulator project:
    [rc_buildingsimulator-jayathissa]:
    [1] Jayathissa, Prageeth, et al. "Optimising building net energy demand with dynamic BIPV shading." Applied Energy 202 (2017): 726-735.
    The implementation of the RC_BuildingSimulator project can be found under the following repository:
    https://github.com/architecture-building-systems/RC_BuildingSimulator

    The RC model (5R1C model) is based on the EN ISO 13790 standard and explains thermal physics with the help of an electrical circuit analogy.
    Principal components of this model are the heat fluxes Phi which are analogous to the electrical current I,
    the thermal conductances H which are the inverse of the thermal resistance and analogous to the electrical conductance G (=1/R),
    the thermal capacitance C analogous to the electrical capacitance C and
    the temperatures T which are analogous to the electrical voltage V.

*** Another paper using the 5R1C model from EN ISO 13790:
    [2] I. Horvat et al. "Dynamic method for calculating energy need in HVAC systems." Transactions of Famena 40 (2016): 47-62.

"""
# clean

# Generic/Built-in
from typing import (
    List,
    Any
)
from functools import (
    lru_cache,
)
import math
from dataclasses import (
    dataclass,
)
from dataclasses_json import (
    dataclass_json,
)
import pvlib
import pandas as pd
import numpy as np

from hisim import (
    dynamic_component,
    utils,
)
from hisim import (
    component as cp,
)
from hisim import (
    loadtypes as lt,
)
from hisim import (
    log,
)
# from hisim.components.configuration import (
#     PhysicsConfig,
# )
# from hisim.components.configuration import (
#     LoadConfig,
# )
from hisim.components.loadprofilegenerator_utsp_connector import (
    UtspLpgConnector,
)
from hisim.simulationparameters import (
    SimulationParameters,
)
from hisim.components.weather import (
    Weather,
)
from hisim.components.loadprofilegenerator_connector import (
    Occupancy,
)


__authors__ = "Vitor Hugo Bellotto Zago"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Dr. Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@lru_cache(maxsize=16)
def calc_solar_heat_gains(
    sun_azimuth,
    direct_normal_irradiance,
    direct_horizontal_irradiance,
    global_horizontal_irradiance,
    direct_normal_irradiance_extra,
    apparent_zenith,
    altitude_tilt,
    azimuth_tilt,
    reduction_factor_with_area,
):
    """Calculates the Solar Gains in the building zone through the set Window.

    :param sun_altitude: Altitude Angle of the Sun in Degrees
    :type sun_altitude: float
    :param sun_azimuth: Azimuth angle of the sun in degrees
    :type sun_azimuth: float
    :param normal_direct_radiation: Normal Direct Radiation from weather file
    :type normal_direct_radiation: float
    :param horizontal_diffuse_radiation: Horizontal Diffuse Radiation from weather file
    :type horizontal_diffuse_radiation: float
    :return: self.incident_solar, Incident Solar Radiation on window
    :return: self.solar_gains - Solar gains in building after transmitting through the window
    :rtype: float
    """
    poa_irrad = pvlib.irradiance.get_total_irradiance(
        altitude_tilt,
        azimuth_tilt,
        apparent_zenith,
        sun_azimuth,
        direct_normal_irradiance,
        global_horizontal_irradiance,
        direct_horizontal_irradiance,
        direct_normal_irradiance_extra,
    )

    if math.isnan(poa_irrad["poa_direct"]):
        return 0

    return poa_irrad["poa_direct"] * reduction_factor_with_area


class BuildingState:

    """BuildingState class."""

    def __init__(
        self,
        thermal_mass_temperature_in_celsius: float,
        thermal_capacitance_in_joule_per_kelvin: float,
    ):
        """Constructs all the neccessary attributes for the BuildingState object."""
        # this is labeled as t_m in the paper [1] (** Check header)
        self.thermal_mass_temperature_in_celsius: float = (
            thermal_mass_temperature_in_celsius
        )
        # this is labeled as c_m in the paper [1] (** Check header)
        self.thermal_capacitance_in_joule_per_kelvin: float = (
            thermal_capacitance_in_joule_per_kelvin
        )

    def calc_stored_thermal_power_in_watt(
        self,
    ) -> float:
        """Calculates the thermal energy stored by the thermal mass per second (thermal power)."""
        return (
            self.thermal_mass_temperature_in_celsius
            * self.thermal_capacitance_in_joule_per_kelvin
        ) / 3600

    def self_copy(
        self,
    ):
        """Copies the Building State."""
        return BuildingState(
            self.thermal_mass_temperature_in_celsius,
            self.thermal_capacitance_in_joule_per_kelvin,
        )


class BuildingControllerState:

    """BuildingControllerState class."""

    def __init__(
        self,
        temperature_building_target_in_celsius: float,
        level_of_utilization: float,
    ):
        """Constructs all the neccessary attributes for the BuildingControllerState object."""
        self.temperature_building_target_in_celsius: float = (
            temperature_building_target_in_celsius
        )
        self.level_of_utilization: float = level_of_utilization

    def clone(self):
        """Copies the BuildingControllerState."""
        return BuildingControllerState(
            temperature_building_target_in_celsius=self.temperature_building_target_in_celsius,
            level_of_utilization=self.level_of_utilization,
        )


@dataclass_json
@dataclass
class BuildingConfig(cp.ConfigBase):

    """Configuration of the Building class."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return Building.get_full_classname()

    name: str
    heating_reference_temperature_in_celsius: float
    building_code: str
    building_heat_capacity_class: str
    initial_internal_temperature_in_celsius: float

    @classmethod
    def get_default_german_single_family_home(
        cls,
    ) -> Any:
        """Gets a default Building."""
        config = BuildingConfig(
            name="Building_1",
            building_code="DE.N.SFH.05.Gen.ReEx.001.002",
            building_heat_capacity_class="medium",
            initial_internal_temperature_in_celsius=23,
            heating_reference_temperature_in_celsius=-14,
        )
        return config


@dataclass_json
@dataclass
class BuildingControllerConfig:

    """Configuration of the Building Controller class."""

    minimal_building_temperature_in_celsius: float
    stop_heating_building_temperature_in_celsius: float


class Building(dynamic_component.DynamicComponent):

    """Building class.

    This class calculates the thermal behaviour of the building based on the RC Simulator (** paper [1]) and the EN ISO 13790 norm (see header).
    The corresponding functions and variables are also described in the paper [2].
    Also it provides multiple typologies of residences based on the EPISCOPE/TABULA project database (* Check header).

    Parameters
    ----------
    building_code :str
        Code reference to a specific residence typology list in EPISCOPE/TABULA database
    building_heat_capacity_class: str
        Heat capacity of residence defined using one of the following terms:
            - very light
            - light
            - medium
            - heavy
            - very heavy
    initial_internal_temperature_in_celsius : float
        Initial internal temperature of residence in Celsius
    sim_params : Simulator
        Simulator object used to carry the simulation using this class

    """

    # Inputs -> heating device
    # either thermal energy delivered from heat pump
    ThermalEnergyDelivered = "ThermalEnergyDelivered"
    # or mass input and temperature input delivered from Thermal Energy Storage (TES)
    MassInput = "MassInput"
    TemperatureInput = "TemperatureInput"

    # Inputs -> occupancy
    HeatingByResidents = "HeatingByResidents"

    # Inputs -> weather
    Altitude = "Altitude"
    Azimuth = "Azimuth"
    ApparentZenith = "ApparentZenith"
    DirectNormalIrradiance = "DirectNormalIrradiance"
    DirectNormalIrradianceExtra = "DirectNormalIrradianceExtra"
    DiffuseHorizontalIrradiance = "DiffuseHorizontalIrradiance"
    GlobalHorizontalIrradiance = "GlobalHorizontalIrradiance"
    TemperatureOutside = "TemperatureOutside"

    # Outputs
    TemperatureMean = "Residence Temperature"
    TemperatureAir = "TemperatureAir"
    TotalEnergyToResidence = "TotalEnergyToResidence"
    SolarGainThroughWindows = "SolarGainThroughWindows"
    StoredEnergyVariation = "StoredEnergyVariation"
    InternalLoss = "InternalLoss"
    OldStoredEnergy = "OldStoredEnergy"
    CurrentStoredEnergy = "CurrentStoredEnergy"
    MassOutput = "MassOutput"
    TemperatureOutput = "TemperatureOutput"
    ReferenceMaxHeatBuildingDemand = "ReferenceMaxHeatBuildingDemand"

    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: BuildingConfig,
    ):
        """Constructs all the neccessary attributes."""
        self.buildingconfig = config
        # dynamic
        self.my_component_inputs: List[dynamic_component.DynamicConnectionInput] = []
        self.my_component_outputs: List[dynamic_component.DynamicConnectionOutput] = []
        super().__init__(
            my_component_inputs=self.my_component_inputs,
            my_component_outputs=self.my_component_outputs,
            name=self.buildingconfig.name,
            my_simulation_parameters=my_simulation_parameters,
        )

        # =================================================================================================================================
        # Initialization of variables

        (self.is_in_cache, self.cache_file_path,) = utils.get_cache_file(
            self.component_name,
            self.buildingconfig,
            self.my_simulation_parameters,
        )
        # labeled as C_m in the paper [1] (** Check header), before c_m
        self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin: float = 0
        self.thermal_capacity_of_building_thermal_mass_reference_in_joule_per_kelvin: float = 0
        # labeled as H_w in the paper [2] (*** Check header), before h_tr_w
        self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin: float
        self.transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin: float
        # labeled as H_tr_em in paper [2] (*** Check header)
        self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin: float = 0
        # labeled as H_tr_ms in paper [2] (*** Check header)
        self.internal_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin: float = 0
        # labeled as h_ms in paper [2] (*** Check header)
        self.heat_transfer_coefficient_between_thermal_mass_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin: float = 0
        # labeled as H_tr_is in paper [2] (** Check header)
        self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_in_watt_per_kelvin: float = 0
        # labeled as h_is in paper [2] (** Check header)
        self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin: float = 0
        # labeled as H_ve in paper [2] (*** Check header), before h_ve_adj
        self.thermal_conductance_by_ventilation_in_watt_per_kelvin: float = 0
        self.thermal_conductance_by_ventilation_reference_in_watt_per_kelvin: float = 0
        # before labeled as a_f
        self.conditioned_floor_area_in_m2: float = 0
        # before labeled as a_m
        self.effective_mass_area_in_m2: float = 0
        # before labeled as a_t
        self.total_internal_surface_area_in_m2: float = 0
        self.room_volume_in_m3: float = 0
        # reference taken from TABULA (* Check header) as Q_ht [kWh/m2.a], before q_ht_ref
        self.total_heat_transfer_reference_in_kilowatthour_per_m2_per_year: float = 0
        # reference taken from TABULA (* Check header) as Q_int [kWh/m2.a], before q_int_ref
        self.internal_heat_sources_reference_in_kilowatthour_per_m2_per_year: float = 0
        # reference taken from TABULA (* Check header) Q_sol [kWh/m2.a], before q_sol_ref (or solar heat sources?)
        self.solar_heat_load_during_heating_seasons_reference_in_kilowatthour_per_m2_per_year: float = 0
        # reference taken from TABULA (* Check header) as Q_H_ind [kWh/m2.a], before q_h_nd_ref
        self.energy_need_for_heating_reference_in_kilowatthour_per_m2_per_year: float = 0
        self.test_new_temperature_in_celsius: float
        self.buildingdata: Any
        self.buildingcode: str
        self.windows: List[Window]
        self.windows_area: float
        self.cache: List[float]
        self.solar_heat_gain_through_windows: List[float]
        # labeled as Phi_ia in paper [1] (** Check header)
        self.heat_flux_indoor_air_in_watt: float
        # labeled as Phi_st in the paper [1] (** Check header)
        self.heat_flux_internal_room_surface_in_watt: float
        # labeled as Phi_m in the paper [1] (** Check header)
        self.heat_flux_thermal_mass_in_watt: float
        self.heat_loss_in_watt: float
        # labeled as Phi_m_tot in the paper [1] (** Check header)
        self.equivalent_heat_flux_in_watt: float
        self.next_thermal_mass_temperature_in_celsius: float

        self.get_building()
        self.max_thermal_building_demand_in_watt = self.calc_max_thermal_building_demand(
            heating_reference_temperature_in_celsius=config.heating_reference_temperature_in_celsius,
            initial_temperature_in_celsius=config.initial_internal_temperature_in_celsius,
        )
        self.build()

        self.state: BuildingState = BuildingState(
            thermal_mass_temperature_in_celsius=config.initial_internal_temperature_in_celsius,
            thermal_capacitance_in_joule_per_kelvin=self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin,
        )
        self.previous_state = self.state.self_copy()

        # =================================================================================================================================
        # Input and Output channels

        self.thermal_power_delivered_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ThermalEnergyDelivered,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            False,
        )
        self.mass_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.MassInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.KG_PER_SEC,
            False,
        )
        self.temperature_input_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            False,
        )

        self.altitude_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.Altitude,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            True,
        )
        self.azimuth_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.Azimuth,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            True,
        )
        self.apparent_zenith_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ApparentZenith,
            lt.LoadTypes.ANY,
            lt.Units.DEGREES,
            True,
        )
        self.direct_normal_irradiance_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DirectNormalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )
        self.direct_normal_irradiance_extra_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DirectNormalIrradianceExtra,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )
        self.direct_horizontal_irradiance_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.DiffuseHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )
        self.global_horizontal_irradiance_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.GlobalHorizontalIrradiance,
            lt.LoadTypes.IRRADIANCE,
            lt.Units.WATT_PER_SQUARE_METER,
            True,
        )

        self.temperature_outside_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.TemperatureOutside,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            True,
        )

        self.occupancy_heat_gain_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.HeatingByResidents,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            True,
        )

        self.thermal_mass_temperature_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.TemperatureMean,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
        )
        self.total_power_to_residence_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.TotalEnergyToResidence,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
        )
        self.solar_gain_through_windows_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.SolarGainThroughWindows,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
        )
        self.var_max_thermal_building_demand_channel: cp.ComponentOutput = (
            self.add_output(
                self.component_name,
                self.ReferenceMaxHeatBuildingDemand,
                lt.LoadTypes.HEATING,
                lt.Units.WATT,
            )
        )

        # =================================================================================================================================
        # Add and get default connections

        self.add_default_connections(
            Weather,
            self.get_weather_default_connections(),
        )
        self.add_default_connections(
            Occupancy,
            self.get_occupancy_default_connections(),
        )
        self.add_default_connections(
            UtspLpgConnector,
            self.get_utsp_default_connections(),
        )

    def get_weather_default_connections(
        self,
    ):
        """Get weather default connnections."""
        log.information("setting weather default connections")
        connections = []
        weather_classname = Weather.get_classname()
        connections.append(
            cp.ComponentConnection(
                Building.Altitude,
                weather_classname,
                Weather.Altitude,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.Azimuth,
                weather_classname,
                Weather.Azimuth,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.ApparentZenith,
                weather_classname,
                Weather.ApparentZenith,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.DirectNormalIrradiance,
                weather_classname,
                Weather.DirectNormalIrradiance,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.DirectNormalIrradianceExtra,
                weather_classname,
                Weather.DirectNormalIrradianceExtra,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.DiffuseHorizontalIrradiance,
                weather_classname,
                Weather.DiffuseHorizontalIrradiance,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.GlobalHorizontalIrradiance,
                weather_classname,
                Weather.GlobalHorizontalIrradiance,
            )
        )
        connections.append(
            cp.ComponentConnection(
                Building.TemperatureOutside,
                weather_classname,
                Weather.TemperatureOutside,
            )
        )
        return connections

    def get_occupancy_default_connections(
        self,
    ):
        """Get occupancy default connections."""
        log.information("setting occupancy default connections")
        connections = []
        occupancy_classname = Occupancy.get_classname()
        connections.append(
            cp.ComponentConnection(
                Building.HeatingByResidents,
                occupancy_classname,
                Occupancy.HeatingByResidents,
            )
        )
        return connections

    def get_utsp_default_connections(
        self,
    ):
        """Gets UTSP default connections."""
        log.information("setting utsp default connections")
        connections = []
        utsp_classname = UtspLpgConnector.get_classname()
        connections.append(
            cp.ComponentConnection(
                Building.HeatingByResidents,
                utsp_classname,
                UtspLpgConnector.HeatingByResidents,
            )
        )
        return connections

    # =================================================================================================================================
    # Simulation of the building class

    def i_simulate(
        self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool
    ) -> None:
        """Simulates the thermal behaviour of the building."""

        # Gets inputs
        if hasattr(self, "solar_gain_through_windows") is False:
            # altitude = stsv.get_input_value(self.altitude_channel)
            azimuth = stsv.get_input_value(self.azimuth_channel)
            direct_normal_irradiance = stsv.get_input_value(self.direct_normal_irradiance_channel)
            direct_horizontal_irradiance = stsv.get_input_value(self.direct_horizontal_irradiance_channel)
            global_horizontal_irradiance = stsv.get_input_value(self.global_horizontal_irradiance_channel)
            direct_normal_irradiance_extra = stsv.get_input_value(self.direct_normal_irradiance_extra_channel)
            apparent_zenith = stsv.get_input_value(self.apparent_zenith_channel)

        occupancy_heat_gain_in_watt = stsv.get_input_value(self.occupancy_heat_gain_channel)
        temperature_outside_in_celsius = stsv.get_input_value(self.temperature_outside_channel)
        thermal_power_delivered_in_watt = stsv.get_input_value(self.thermal_power_delivered_channel)

        # # With Thermal Energy Storage (TES) [In Development]
        # if self.mass_input_channel.source_output is not None:
        #     if force_convergence:
        #         return

        #     thermal_power_delivered_in_watt = stsv.get_input_value(self.thermal_power_delivered_channel)
        #     mass_input_in_kilogram_per_second = stsv.get_input_value(self.mass_input_channel)

        #     temperature_input_in_celsius = stsv.get_input_value(self.temperature_input_channel)

        #     if thermal_power_delivered_in_watt > 0 and (
        #         mass_input_in_kilogram_per_second == 0
        #         and temperature_input_in_celsius == 0
        #     ):
        #         """first iteration --> random numbers"""
        #         temperature_input_in_celsius = 40.456
        #         mass_input_in_kilogram_per_second = 0.0123

        #     if thermal_power_delivered_in_watt > 0:

        #         massflows_possible_in_kilogram_per_second = (
        #             LoadConfig.possible_massflows_load
        #         )
        #         mass_flow_level = 0
        #         # K = W / (J/kgK * kg/s); delta T in Kelvin = delta T in Celsius; heat capacity J/kgK = J/kg°C
        #         temperature_delta_heat_in_kelvin = thermal_power_delivered_in_watt / (
        #             PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
        #             * massflows_possible_in_kilogram_per_second[mass_flow_level]
        #         )
        #         while temperature_delta_heat_in_kelvin > LoadConfig.delta_T:
        #             mass_flow_level += 1
        #             temperature_delta_heat_in_kelvin = thermal_power_delivered_in_watt / (
        #                 PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
        #                 * massflows_possible_in_kilogram_per_second[mass_flow_level]
        #             )

        #         mass_input_load_in_kilogram_per_timestep = (
        #             massflows_possible_in_kilogram_per_second[mass_flow_level]
        #             * self.seconds_per_timestep
        #         )

        #         energy_demand_in_joule_per_timestep = (
        #             thermal_power_delivered_in_watt * self.seconds_per_timestep
        #         )
        #         enthalpy_slice_in_joule_per_timestep = (
        #             mass_input_load_in_kilogram_per_timestep
        #             * temperature_input_in_celsius
        #             * PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
        #         )
        #         enthalpy_new_in_joule_per_timestep = (
        #             enthalpy_slice_in_joule_per_timestep
        #             - energy_demand_in_joule_per_timestep
        #         )
        #         temperature_new_in_celsius = enthalpy_new_in_joule_per_timestep / (
        #             mass_input_load_in_kilogram_per_timestep
        #             * PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
        #         )

        #     else:
        #         # no water is flowing
        #         temperature_new_in_celsius = temperature_input_in_celsius
        #         mass_input_load_in_kilogram_per_timestep = 0

        #     self.test_new_temperature_in_celsius = temperature_new_in_celsius

        # # Only with HeatPump
        # elif self.thermal_power_delivered_channel.source_output is not None:
        #     thermal_power_delivered_in_watt = stsv.get_input_value(self.thermal_power_delivered_channel)
        # else:
        #     thermal_power_delivered_in_watt = sum(
        #         self.get_dynamic_inputs(
        #             stsv=stsv, tags=[lt.InandOutputType.HEAT_TO_BUILDING]
        #         )
        #     )

        previous_thermal_mass_temperature_in_celsius = (
            self.state.thermal_mass_temperature_in_celsius
        )

        # Performs calculations
        if hasattr(self, "solar_gain_through_windows") is False:
            # @JG I guess you wanted to transfer W to Wh
            solar_heat_gain_through_windows = self.get_solar_heat_gain_through_windows(
                # altitude=altitude,
                azimuth=azimuth,
                direct_normal_irradiance=direct_normal_irradiance,
                direct_horizontal_irradiance=direct_horizontal_irradiance,
                global_horizontal_irradiance=global_horizontal_irradiance,
                direct_normal_irradiance_extra=direct_normal_irradiance_extra,
                apparent_zenith=apparent_zenith,
            )
        else:
            solar_heat_gain_through_windows = self.solar_heat_gain_through_windows[
                timestep
            ]

        (
            thermal_mass_average_bulk_temperature_in_celsius,
            heat_loss_in_watt,
        ) = self.calc_crank_nicolson(
            thermal_power_delivered_in_watt=thermal_power_delivered_in_watt,
            internal_heat_gains_in_watt=occupancy_heat_gain_in_watt,
            solar_heat_gains_in_watt=solar_heat_gain_through_windows,
            outside_temperature_in_celsius=temperature_outside_in_celsius,
            thermal_mass_temperature_prev_in_celsius=previous_thermal_mass_temperature_in_celsius,
        )
        self.state.thermal_mass_temperature_in_celsius = thermal_mass_average_bulk_temperature_in_celsius

        # Returns outputs

        stsv.set_output_value(self.thermal_mass_temperature_channel, thermal_mass_average_bulk_temperature_in_celsius)
        # stsv.set_output_value(self.t_airC, t_air)
        # phi_loss is already given in W, time correction factor applied to thermal transmittance h_tr
        stsv.set_output_value(self.total_power_to_residence_channel, heat_loss_in_watt)
        stsv.set_output_value(self.solar_gain_through_windows_channel, solar_heat_gain_through_windows)
        stsv.set_output_value(self.var_max_thermal_building_demand_channel, self.max_thermal_building_demand_in_watt)

        # Saves solar gains cache
        if not self.is_in_cache:
            self.cache[timestep] = solar_heat_gain_through_windows
            if timestep + 1 == self.my_simulation_parameters.timesteps:
                database = pd.DataFrame(
                    self.cache,
                    columns=["solar_gain_through_windows"],
                )
                database.to_csv(
                    self.cache_file_path,
                    sep=",",
                    decimal=".",
                    index=False,
                )

    # =================================================================================================================================

    def i_save_state(
        self,
    ) -> None:
        """Saves the current state."""
        self.previous_state = self.state.self_copy()

    def i_prepare_simulation(
        self,
    ) -> None:
        """Prepares the simulation."""
        pass

    def i_restore_state(
        self,
    ) -> None:
        """Restores the previous state."""
        self.state = self.previous_state.self_copy()

    def i_doublecheck(
        self,
        timestep: int,
        stsv: cp.SingleTimeStepValues,
    ) -> None:
        """Doublechecks."""
        pass

    def build(
        self,
    ):
        """Build function.

        The function sets important constants and parameters for the calculations.
        It imports the building dataset from TABULA and gets physical parameters and thermal conductances.
        """

        self.seconds_per_timestep = self.my_simulation_parameters.seconds_per_timestep
        self.timesteps = self.my_simulation_parameters.timesteps
        self.parameters = [
            self.building_heat_capacity_class,
            self.buildingcode,
        ]

        # CONSTANTS
        # Heat transfer coefficient between nodes "m" and "s" (12.2.2 E64 P79); labeled as h_ms in paper [2] (*** Check header)
        self.heat_transfer_coefficient_between_thermal_mass_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin = 9.1
        # Dimensionless ratio between surfaces and the useful surfaces (7.2.2.2 E9 P36); labeled as A_at in paper [2] (*** Check header); before lambda_at
        self.ratio_between_internal_surface_area_and_floor_area = 4.5
        # Heat transfer coefficient between nodes "air" and "s" (7.2.2.2 E9 P35); labeled as h_is in paper [2] (*** Check header)
        self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin = 3.45

        self.building_heat_capacity_class_f_a = {
            "very light": 2.5,
            "light": 2.5,
            "medium": 2.5,
            "heavy": 3.0,
            "very heavy": 3.5,
        }
        self.building_heat_capacity_class_f_c = {
            "very light": 8e4,
            "light": 1.1e5,
            "medium": 1.65e5,
            "heavy": 2.6e5,
            "very heavy": 3.7e5,
        }

        self.ven_method = "EPISCOPE"

        # Get physical parameters
        self.get_physical_param()

        # Gets conductances
        self.get_conductances()

    def calc_max_thermal_building_demand(
        self,
        initial_temperature_in_celsius: float,
        heating_reference_temperature_in_celsius: float,
    ) -> Any:
        """Calculate maximal thermal building demand using TABULA data."""

        vals1_in_watt_per_m2_per_kelvin = float(self.buildingdata["h_Transmission"].values[0])

        if vals1_in_watt_per_m2_per_kelvin is None:
            raise ValueError("h_Transmission was none.")
        vals2_in_watt_per_m2_per_kelvin = float(self.buildingdata["h_Ventilation"].values[0])
        conditioned_floor_area_in_m2 = float(self.buildingdata["A_C_Ref"].values[0])

        # dQ/dt = h * (T2-T1) * A -> [W]
        max_thermal_building_demand_in_watt = (
            (vals1_in_watt_per_m2_per_kelvin + vals2_in_watt_per_m2_per_kelvin)
            * (
                initial_temperature_in_celsius
                - heating_reference_temperature_in_celsius
            )
            * conditioned_floor_area_in_m2
        )
        return max_thermal_building_demand_in_watt

    def __str__(
        self,
    ):
        """Return lines from report as string format."""
        entire = str()
        lines = self.write_to_report()
        for (
            index,
            line,
        ) in enumerate(lines):
            if index == 0:
                entire = line
            else:
                entire = f"{entire}\n{line}"
        return entire

    def write_to_report(
        self,
    ):
        """Writes important variables to a report."""
        lines = []
        lines.append(f"Name: {self.component_name}")
        lines.append(f"Code: {self.buildingcode}")

        lines.append("")
        lines.append("Conductances:")
        lines.append(
            f"H_tr_w [W/K]: {self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin:4.2f}"
        )
        lines.append(
            f"H_tr_em [W/K]: {self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin:4.2f}"
        )
        lines.append(
            f"H_tr_is [W/K]: {self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_in_watt_per_kelvin:4.2f}"
        )
        lines.append(
            f"H_tr_ms [W/K]: {self.internal_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin:4.2f}"
        )
        lines.append(
            f"H_ve_adj [W/K]: {self.thermal_conductance_by_ventilation_in_watt_per_kelvin:4.2f}"
        )
        lines.append(
            f"H_Ventilation [kWh/a]: {self.thermal_conductance_by_ventilation_reference_in_watt_per_kelvin}"
        )

        lines.append(" ")
        lines.append("Areas:")
        lines.append(f"A_f [m^2]: {self.conditioned_floor_area_in_m2:4.1f}")
        lines.append(f"A_m [m^2]: {self.effective_mass_area_in_m2:4.1f}")
        lines.append(f"A_t [m^2]: {self.total_internal_surface_area_in_m2:4.1f}")
        lines.append(f"Room volume [m^3]: {self.room_volume_in_m3}")

        lines.append(" ")
        lines.append("Capacitance:")
        lines.append(
            f"Capacitance [Wh/m^2.K]: {(self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin * 3600 / self.conditioned_floor_area_in_m2):4.2f}"
        )
        lines.append(
            f"Capacitance [Wh/K]: {(self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin * 3600):4.2f}"
        )
        lines.append(
            f"Capacitance [J/K]: {self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin:4.2f}"
        )

        lines.append(
            f"Capacitance Ref [Wh/m^2.K]: {self.thermal_capacity_of_building_thermal_mass_reference_in_joule_per_kelvin:4.2f}"
        )
        lines.append(
            f"Capacitance Ref [Wh/K]: {(self.thermal_capacity_of_building_thermal_mass_reference_in_joule_per_kelvin * self.conditioned_floor_area_in_m2):4.2f}"
        )
        lines.append(
            f"Capacitance Ref [J/K]: {(self.thermal_capacity_of_building_thermal_mass_reference_in_joule_per_kelvin * self.conditioned_floor_area_in_m2 / 3600):4.2f}"
        )

        lines.append(" ")
        lines.append("Heat Transfers:")
        lines.append(
            f"Annual heating losses Q_ht [kWh/m^2.a]: {self.total_heat_transfer_reference_in_kilowatthour_per_m2_per_year}"
        )
        lines.append(
            f"Annual heating losses Q_ht [kWh/a]: {self.total_heat_transfer_reference_in_kilowatthour_per_m2_per_year * self.conditioned_floor_area_in_m2}"
        )
        lines.append(
            f"Q_int [kWh/m^2.a]: {self.internal_heat_sources_reference_in_kilowatthour_per_m2_per_year}"
        )
        lines.append(
            f"Q_int [kWh/a]: {self.internal_heat_sources_reference_in_kilowatthour_per_m2_per_year * self.conditioned_floor_area_in_m2}"
        )
        lines.append(
            f"Q_sol [kWh/m^2.a]: {self.solar_heat_load_during_heating_seasons_reference_in_kilowatthour_per_m2_per_year}"
        )
        lines.append(
            f"Q_sol [kWh/a]: {self.solar_heat_load_during_heating_seasons_reference_in_kilowatthour_per_m2_per_year * self.conditioned_floor_area_in_m2}"
        )
        lines.append("=============== REFERENCE ===============")
        lines.append(
            f"Balance Heating Demand Reference [kWh/m^2.a]: {self.energy_need_for_heating_reference_in_kilowatthour_per_m2_per_year}"
        )
        lines.append(
            f"Balance Heating Demand Reference [kWh/a]: {self.energy_need_for_heating_reference_in_kilowatthour_per_m2_per_year * self.conditioned_floor_area_in_m2}"
        )
        return lines

    # =====================================================================================================================================
    # Calculation of the heat transfer coefficients or thermal conductances.
    # (**/*** Check header)

    @property
    def transmission_heat_transfer_coeffcient_1_in_watt_per_kelvin(
        self,
    ):
        """Definition to simplify calc_phi_m_tot. Long form for H_tr_1.

        # (C.6) in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        return 1.0 / (
            1.0 / self.thermal_conductance_by_ventilation_in_watt_per_kelvin
            + 1.0
            / self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_in_watt_per_kelvin
        )

    @property
    def transmission_heat_transfer_coefficient_2_in_watt_per_kelvin(
        self,
    ):
        """Definition to simplify calc_phi_m_tot. Long form for H_tr_2.

        # (C.7) in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        return (
            self.transmission_heat_transfer_coeffcient_1_in_watt_per_kelvin
            + self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
        )

    @property
    def transmission_heat_transfer_coefficient_3_in_watt_per_kelvin(
        self,
    ):
        """Definition to simplify calc_phi_m_tot. Long form for H_tr_3.

        # (C.8) in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        return 1.0 / (
            1.0 / self.transmission_heat_transfer_coefficient_2_in_watt_per_kelvin
            + 1.0
            / self.internal_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
        )

    def get_thermal_conductance_between_exterior_and_windows_and_door_in_watt_per_kelvin(
        self,
    ):
        """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
        # Long from for H_tr_w: Conductance between exterior temperature and surface temperature
        # Objects: Doors, windows, curtain walls and windowed walls ISO 7.2.2.2
        w_s = [
            "Window_1",
            "Window_2",
            "Door_1",
        ]

        self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin = 0.0
        for w_i in w_s:
            self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin += float(
                self.buildingdata["H_Transmission_" + w_i].values[0]
            )

    def get_thermal_conductance_between_thermal_mass_and_internal_surface_in_watt_per_kelvin(
        self,
    ):
        """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
        # Long from for H_tr_ms, this is the same as internal pasrt of transmission heat transfer coefficient for opaque elements
        self.internal_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin = (
            self.effective_mass_area_in_m2
            * self.heat_transfer_coefficient_between_thermal_mass_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
        )

    def get_thermal_conductance_of_opaque_surfaces_in_watt_per_kelvin(
        self,
    ):
        """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
        # Long from for H_tr_op: H_tr_op = 1/ (1/H_tr_ms + 1/H_tr_em) with
        # H_tr_ms: Conductance of opaque surfaces to interior [W/K] and H_tr_em: Conductance of opaque surfaces to exterior [W/K]
        opaque_walls = [
            "Wall_1",
            "Wall_2",
            "Wall_3",
            "Roof_1",
            "Roof_2",
            "Floor_1",
            "Floor_2",
        ]

        # Version 1
        # self.h_tr_em = 0.0
        # for ow in opaque_walls:
        #    self.h_tr_em += float(self.buildingdata["H_Transmission_" + ow].values[0])

        # Version 2
        self.transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin = 0.0
        for o_w in opaque_walls:
            self.transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin += float(
                self.buildingdata["H_Transmission_" + o_w].values[0]
            )

        self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin = 1 / (
            (
                1
                / self.transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
            )
            - (
                1
                / self.internal_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
            )
        )

    def get_thermal_conductance_between_indoor_air_and_internal_surface_in_watt_per_kelvin(
        self,
    ):
        """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
        # Long from for H_tr_is: Conductance between air temperature and surface temperature
        self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_in_watt_per_kelvin = (
            self.total_internal_surface_area_in_m2
            * self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
        )

    def get_thermal_conductance_ventilation_in_watt_per_kelvin(
        self,
    ):
        """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
        # Long from for H_ve_adj: Ventilation
        if self.ven_method == "RC_BuildingSimulator":
            """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""
            air_changes_per_hour_through_ventilation = 1.5
            air_changes_per_hour_through_infiltration = 0.5
            ventilation_efficiency = 0.6
            # Determine the ventilation conductance
            total_air_changes_per_hour = (
                air_changes_per_hour_through_infiltration
                + air_changes_per_hour_through_ventilation
            )
            # temperature adjustment factor taking ventilation and infiltration
            # [ISO: E -27]
            b_ek = (
                1
                - (
                    air_changes_per_hour_through_ventilation
                    / total_air_changes_per_hour
                )
                * ventilation_efficiency
            )
            self.thermal_conductance_by_ventilation_in_watt_per_kelvin = float(
                1200
                * b_ek
                * self.room_volume_in_m3
                * (total_air_changes_per_hour / 3600)
            )  # Conductance through ventilation [W/M]
        elif self.ven_method == "EPISCOPE":
            # cp = 0.00028378 * 1E3  # [Wh/m3K]
            c_p = 0.34
            self.thermal_conductance_by_ventilation_in_watt_per_kelvin = (
                c_p
                * float(
                    self.buildingdata["n_air_use"]
                    + self.buildingdata["n_air_infiltration"]
                )
                * self.conditioned_floor_area_in_m2
                * float(self.buildingdata["h_room"])
            )

    def get_conductances(
        self,
    ):
        """Get the thermal conductances based on the norm EN ISO 13970.

        :key
        """
        self.get_thermal_conductance_between_exterior_and_windows_and_door_in_watt_per_kelvin()
        self.get_thermal_conductance_between_thermal_mass_and_internal_surface_in_watt_per_kelvin()
        self.get_thermal_conductance_of_opaque_surfaces_in_watt_per_kelvin()
        self.get_thermal_conductance_between_indoor_air_and_internal_surface_in_watt_per_kelvin()
        self.get_thermal_conductance_ventilation_in_watt_per_kelvin()

    # =====================================================================================================================================
    # Get building parameters.
    # (* Check header)

    def get_physical_param(
        self,
    ):
        """Get the physical parameters from the building data."""
        # Windows area
        self.get_windows()

        # Reference area [m^2] (TABULA: Reference floor area )Ref: ISO standard 7.2.2.2
        self.conditioned_floor_area_in_m2 = float(
            self.buildingdata["A_C_Ref"].values[0]
        )
        # total_internal_area = buildingdata["A_Estim_Floor"][1]

        self.effective_mass_area_in_m2 = (
            self.conditioned_floor_area_in_m2
            * self.building_heat_capacity_class_f_a[self.building_heat_capacity_class]
        )
        self.total_internal_surface_area_in_m2 = (
            self.conditioned_floor_area_in_m2
            * self.ratio_between_internal_surface_area_and_floor_area
        )

        # Room Capacitance [J/K] (TABULA: Internal heat capacity) Ref: ISO standard 12.3.1.2
        self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin = (
            self.building_heat_capacity_class_f_c[self.building_heat_capacity_class]
            * self.conditioned_floor_area_in_m2
        )

        # Building volume (TABULA: Conditioned building volume)
        self.room_volume_in_m3 = float(self.buildingdata["V_C"].values[0])

        # Reference properties from TABULA, but not used in the model
        # Floor area related heat load during heating season
        self.solar_heat_load_during_heating_seasons_reference_in_kilowatthour_per_m2_per_year = float(
            (self.buildingdata["q_sol"].values[0])
        )
        # Floor area related internal heat sources during heating season
        self.internal_heat_sources_reference_in_kilowatthour_per_m2_per_year = float(
            self.buildingdata["q_int"].values[0]
        )
        # Floor area related annual losses
        self.total_heat_transfer_reference_in_kilowatthour_per_m2_per_year = float(
            self.buildingdata["q_ht"].values[0]
        )
        # Energy need for heating
        self.energy_need_for_heating_reference_in_kilowatthour_per_m2_per_year = float(
            self.buildingdata["q_h_nd"].values[0]
        )
        # Internal heat capacity per m2 reference area [Wh/(m^2.K)] (TABULA: Internal heat capacity)
        self.thermal_capacity_of_building_thermal_mass_reference_in_joule_per_kelvin = (
            float(self.buildingdata["c_m"].values[0])
        )

        # Heat transfer coefficient by ventilation
        self.thermal_conductance_by_ventilation_reference_in_watt_per_kelvin = (
            float(self.buildingdata["h_Ventilation"].values[0])
            * self.conditioned_floor_area_in_m2
        )

    def get_building(
        self,
    ):
        """Get the building code from a TABULA building."""
        d_f = pd.read_csv(
            utils.HISIMPATH["housing"],
            decimal=",",
            sep=";",
            encoding="cp1252",
            low_memory=False,
        )

        # Gets parameters from chosen building
        self.buildingdata = d_f.loc[d_f["Code_BuildingVariant"] == self.buildingconfig.building_code]
        self.buildingcode = self.buildingconfig.building_code
        self.building_heat_capacity_class = self.buildingconfig.building_heat_capacity_class

    def get_windows(
        self,
    ):
        """Retrieves data about windows sizes.

        :return:
        """

        self.windows = []
        self.windows_area = 0.0
        south_angle = 180

        windows_angles = {
            "South": south_angle,
            "East": south_angle - 90,
            "North": south_angle - 180,
            "West": south_angle + 90,
        }

        windows_directions = [
            "South",
            "East",
            "North",
            "West",
        ]
        f_w = self.buildingdata["F_w"].values[0]
        f_f = self.buildingdata["F_f"].values[0]
        f_sh_vertical = self.buildingdata["F_sh_vert"].values[0]
        g_gln = self.buildingdata["g_gl_n"].values[0]
        for windows_direction in windows_directions:
            area = float(self.buildingdata["A_Window_" + windows_direction])
            if area != 0.0:
                self.windows.append(
                    Window(
                        azimuth_tilt=windows_angles[windows_direction],
                        area=area,
                        frame_area_fraction_reduction_factor=f_f,
                        glass_solar_transmittance=g_gln,
                        nonperpendicular_reduction_factor=f_w,
                        external_shading_vertical_reduction_factor=f_sh_vertical,
                    )
                )
                self.windows_area += area
        # if nothing exists, initialize the empty arrays for caching, else read stuff
        if (
            not self.is_in_cache
        ):  # cache_filepath is None or  (not os.path.isfile(cache_filepath)):
            self.cache = [0] * self.my_simulation_parameters.timesteps
        else:
            self.solar_heat_gain_through_windows = pd.read_csv(
                self.cache_file_path,
                sep=",",
                decimal=".",
            )["solar_gain_through_windows"].tolist()

    # =====================================================================================================================================
    # Calculate solar heat gain through windows.
    # (** Check header)

    # @cached(cache=LRUCache(maxsize=16))
    def get_solar_heat_gain_through_windows(
        self,
        azimuth,
        direct_normal_irradiance,
        direct_horizontal_irradiance,
        global_horizontal_irradiance,
        direct_normal_irradiance_extra,
        apparent_zenith,
    ):  # altitude,
        """Calculates the thermal solar gain passed to the building through the windows.

        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        solar_heat_gains = 0.0

        if (
            direct_normal_irradiance != 0
            or direct_horizontal_irradiance != 0
            or global_horizontal_irradiance != 0
        ):

            for window in self.windows:
                solar_heat_gain = calc_solar_heat_gains(
                    sun_azimuth=azimuth,
                    direct_normal_irradiance=direct_normal_irradiance,
                    direct_horizontal_irradiance=direct_horizontal_irradiance,
                    global_horizontal_irradiance=global_horizontal_irradiance,
                    direct_normal_irradiance_extra=direct_normal_irradiance_extra,
                    apparent_zenith=apparent_zenith,
                    altitude_tilt=window.altitude_tilt,
                    azimuth_tilt=window.azimuth_tilt,
                    reduction_factor_with_area=window.reduction_factor_with_area,
                )
                solar_heat_gains += solar_heat_gain
        return solar_heat_gains

    # =====================================================================================================================================
    # Calculation of the heat flows from internal and solar heat sources
    # (**/*** Check header)

    def calc_heat_flow(
        self,
        # this is labeled as Phi_int in paper [1] (** Check header)
        internal_heat_gains_in_watt,
        # this is labeled as Phi_sol in paper [1] (** Check header)
        solar_heat_gains_in_watt,
        # power_demand_in_watt,
    ):
        """Calculates the heat flow from the solar gains, heating/cooling system, and internal gains into the building.

        The input of the building is split into the air node, surface node, and thermal mass node based on
        on the following equations

        #C.1 - C.3 in [C.3 ISO 13790]

        Note that this equation has diverged slightly from the standard
        as the heating/cooling node can enter any node depending on the
        emission system selected
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """

        # Calculates the heat flows to various points of the building based on the breakdown in section C.2, formulas C.1-C.3
        # Heat flow to the air node in W, before labeled Phi_ia
        # self.heat_flux_indoor_air_in_watt = (
        #     0.5 * internal_heat_gains_in_watt + power_demand_in_watt
        # )
        self.heat_flux_indoor_air_in_watt = 0.5 * internal_heat_gains_in_watt
        # Heat flow to the surface node in W, before labeled Phi_st
        self.heat_flux_internal_room_surface_in_watt = (
            1
            - (self.effective_mass_area_in_m2 / self.total_internal_surface_area_in_m2)
            - (
                self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
                / (9.1 * self.total_internal_surface_area_in_m2)
            )
        ) * (0.5 * internal_heat_gains_in_watt + solar_heat_gains_in_watt)

        # Heat flow to the thermal mass node in W, before labeled Phi_m
        self.heat_flux_thermal_mass_in_watt = (
            self.effective_mass_area_in_m2 / self.total_internal_surface_area_in_m2
        ) * (0.5 * internal_heat_gains_in_watt + solar_heat_gains_in_watt)

        # Heat loss in W, before labeled Phi_loss
        self.heat_loss_in_watt = (
            self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
            / (9.1 * self.total_internal_surface_area_in_m2)
        ) * (0.5 * internal_heat_gains_in_watt + solar_heat_gains_in_watt)
        return self.heat_loss_in_watt

    # =====================================================================================================================================
    # Determination of different temperatures T_air, T_s, T_m,t and T_m and global heat transfer Phi_m_tot which are used in crank-nicolson method
    # (**/*** Check header)

    def calc_next_thermal_mass_temperature_in_celsius(
        self,
        previous_thermal_mass_temperature_in_celsius,
    ):
        """Primary Equation, calculates the temperature of the next time step: T_m,t.

        # (C.4) in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        self.next_thermal_mass_temperature_in_celsius = (
            (
                previous_thermal_mass_temperature_in_celsius
                * (
                    (
                        self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin
                        / self.seconds_per_timestep
                    )
                    - 0.5
                    * (
                        self.transmission_heat_transfer_coefficient_3_in_watt_per_kelvin
                        + self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
                    )
                )
            )
            + self.equivalent_heat_flux_in_watt
        ) / (
            (
                self.thermal_capacity_of_building_thermal_mass_in_joule_per_kelvin
                / self.seconds_per_timestep
            )
            + 0.5
            * (
                self.transmission_heat_transfer_coefficient_3_in_watt_per_kelvin
                + self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
            )
        )

    def calc_equivalent_heat_flux_in_watt(self, temperature_outside_in_celsius, thermal_power_delivered_in_watt):
        """Calculates a global heat transfer: Phi_m_tot.

        This is a definition used to simplify equation calc_t_m_next so it's not so long to write out
        # (C.5) in [C.3 ISO 13790]
        # h_ve = h_ve_adj and t_supply = t_out [9.3.2 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        # ASSUMPTION: Supply air comes straight from the outside air
        # here Phi_HC,nd is not heating or cooling demand but thermal power delivered
        t_supply = temperature_outside_in_celsius

        self.equivalent_heat_flux_in_watt = (
            self.heat_flux_thermal_mass_in_watt
            + self.external_part_of_transmission_heat_transfer_coefficient_for_opaque_elements_in_watt_per_kelvin
            * temperature_outside_in_celsius
            + self.transmission_heat_transfer_coefficient_3_in_watt_per_kelvin
            * (
                self.heat_flux_internal_room_surface_in_watt
                + self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
                * temperature_outside_in_celsius
                + self.transmission_heat_transfer_coeffcient_1_in_watt_per_kelvin
                * (
                    (
                        (self.heat_flux_indoor_air_in_watt + thermal_power_delivered_in_watt)
                        / self.thermal_conductance_by_ventilation_in_watt_per_kelvin
                    )
                    + t_supply
                )
            )
            / self.transmission_heat_transfer_coefficient_2_in_watt_per_kelvin
        )

    def calc_thermal_mass_averag_bulk_temperature_in_celsius_used_for_calculations(
        self,
        previous_thermal_mass_temperature_in_celsius,
    ):
        """Temperature used for the calculations, average between newly calculated and previous bulk temperature: T_m.

        # (C.9) in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        return (
            previous_thermal_mass_temperature_in_celsius
            + self.next_thermal_mass_temperature_in_celsius
        ) / 2

    def calc_temperature_of_internal_room_surfaces_in_celsius(
        self,
        temperature_outside_in_celsius,
        thermal_mass_temperature_in_celsius,
        thermal_power_delivered_in_watt
    ):
        """Calculate the temperature of the inside room surfaces: T_s.

        # (C.10) in [C.3 ISO 13790]
        # h_ve = h_ve_adj and t_supply = t_out [9.3.2 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        # ASSUMPTION: Supply air comes straight from the outside air
        # here Phi_HC,nd is not heating or cooling demand but thermal power delivered
        t_supply = temperature_outside_in_celsius

        return (
            self.heat_transfer_coefficient_between_thermal_mass_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
            * thermal_mass_temperature_in_celsius
            + self.heat_flux_internal_room_surface_in_watt
            + self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
            * temperature_outside_in_celsius
            + self.transmission_heat_transfer_coeffcient_1_in_watt_per_kelvin
            * (
                t_supply
                + (self.heat_flux_indoor_air_in_watt + thermal_power_delivered_in_watt)
                / self.thermal_conductance_by_ventilation_in_watt_per_kelvin
            )
        ) / (
            self.heat_transfer_coefficient_between_thermal_mass_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
            + self.transmission_heat_transfer_coefficient_for_windows_and_door_in_watt_per_kelvin
            + self.transmission_heat_transfer_coeffcient_1_in_watt_per_kelvin
        )

    def calc_temperature_of_the_inside_air_in_celsius(
        self,
        temperature_outside_in_celsius,
        temperature_internal_room_surfaces_in_celsius,
        thermal_power_delivered_in_watt
    ):
        """Calculate the temperature of the air node: T_air.

        # (C.11) in [C.3 ISO 13790]
        # h_ve = h_ve_adj and t_supply = t_out [9.3.2 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        # ASSUMPTION: Supply air comes straight from the outside air
        # here Phi_HC,nd is not heating or cooling demand but thermal power delivered
        t_supply = temperature_outside_in_celsius

        return (
            self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
            * temperature_internal_room_surfaces_in_celsius
            + self.thermal_conductance_by_ventilation_in_watt_per_kelvin * t_supply
            + thermal_power_delivered_in_watt
            + self.heat_flux_indoor_air_in_watt
        ) / (
            self.heat_transfer_coefficient_between_indoor_air_and_internal_surface_with_fixed_value_in_watt_per_m2_per_kelvin
            + self.thermal_conductance_by_ventilation_in_watt_per_kelvin
        )

    def calc_crank_nicolson(
        self,
        internal_heat_gains_in_watt,
        solar_heat_gains_in_watt,
        outside_temperature_in_celsius,
        thermal_mass_temperature_prev_in_celsius,
        thermal_power_delivered_in_watt
    ):
        """Determines node temperatures and computes derivation to determine the new node temperatures.

        Used in: has_demand(), solve_energy(), calc_energy_demand()
        # section C.3 in [C.3 ISO 13790]
        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        Alternatively, described in paper [2].
        """

        # Updates flows
        heat_loss_in_watt = self.calc_heat_flow(
            internal_heat_gains_in_watt,
            solar_heat_gains_in_watt,
        )

        # Updates total flow
        self.calc_equivalent_heat_flux_in_watt(outside_temperature_in_celsius, thermal_power_delivered_in_watt)

        # calculates the new bulk temperature POINT from the old one # CHECKED Requires t_m_prev
        self.calc_next_thermal_mass_temperature_in_celsius(
            thermal_mass_temperature_prev_in_celsius
        )

        # calculates the AVERAGE bulk temperature used for the remaining
        thermal_mass_average_bulk_temperature_in_celsius = (
            self.calc_thermal_mass_averag_bulk_temperature_in_celsius_used_for_calculations(
                thermal_mass_temperature_prev_in_celsius
            )
        )

        # # Updates internal surface temperature (t_s)
        # internal_room_surface_temperature_in_celsius = (
        #     self.calc_temperature_of_internal_room_surfaces_in_celsius(
        #         outside_temperature_in_celsius,
        #         thermal_mass_average_bulk_temperature_in_celsius,
        #     )
        # )

        # # Updates indoor air temperature (t_air)
        # indoor_air_temperature_in_celsius = (
        #     self.calc_temperature_of_the_inside_air_in_celsius(
        #         outside_temperature_in_celsius,
        #         internal_room_surface_temperature_in_celsius,
        #     )
        # )

        return (
            thermal_mass_average_bulk_temperature_in_celsius,
            heat_loss_in_watt,
        )
        # return t_m, t_air, t_s, indoor_air_temperature_in_celsius,internal_room_surface_temperature_in_celsius,


class Window:

    """Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)."""

    def __init__(
        self,
        azimuth_tilt=None,
        altitude_tilt=90,
        area=None,
        glass_solar_transmittance=0.6,
        frame_area_fraction_reduction_factor=0.3,
        external_shading_vertical_reduction_factor=1.0,
        nonperpendicular_reduction_factor=0.9,
    ):
        """Constructs all the neccessary attributes."""
        # Angles
        self.altitude_tilt = altitude_tilt
        self.azimuth_tilt = azimuth_tilt
        self.altitude_tilt_rad = math.radians(altitude_tilt)
        self.azimuth_tilt_rad = math.radians(azimuth_tilt)

        # Area
        self.area = area

        # Transmittance
        self.glass_solar_transmittance = glass_solar_transmittance

        # Incident Solar Radiation
        self.incident_solar: int

        # Reduction factors
        self.nonperpendicular_reduction_factor = nonperpendicular_reduction_factor
        self.external_shading_vertical_reduction_factor = (
            external_shading_vertical_reduction_factor
        )
        self.frame_area_fraction_reduction_factor = frame_area_fraction_reduction_factor

        self.reduction_factor = (
            glass_solar_transmittance
            * nonperpendicular_reduction_factor
            * external_shading_vertical_reduction_factor
            * (1 - frame_area_fraction_reduction_factor)
        )

        self.reduction_factor_with_area = self.reduction_factor * area

    # @cached(cache=LRUCache(maxsize=5))
    # @lru_cache
    def calc_solar_gains(
        self,
        sun_azimuth,
        direct_normal_irradiance,
        direct_horizontal_irradiance,
        global_horizontal_irradiance,
        direct_normal_irradiance_extra,
        apparent_zenith,
    ):
        """Calculates the Solar Gains in the building zone through the set Window.

        :param sun_altitude: Altitude Angle of the Sun in Degrees
        :type sun_altitude: float
        :param sun_azimuth: Azimuth angle of the sun in degrees
        :type sun_azimuth: float
        :param normal_direct_radiation: Normal Direct Radiation from weather file
        :type normal_direct_radiation: float
        :param horizontal_diffuse_radiation: Horizontal Diffuse Radiation from weather file
        :type horizontal_diffuse_radiation: float
        :return: self.incident_solar, Incident Solar Radiation on window
        :return: self.solar_gains - Solar gains in building after transmitting through the window
        :rtype: float
        """
        albedo = 0.4
        # automatic pd time series in future pvlib version
        # calculate airmass
        airmass = pvlib.atmosphere.get_relative_airmass(apparent_zenith)
        # use perez model to calculate the plane of array diffuse sky radiation
        poa_sky_diffuse = pvlib.irradiance.perez(
            self.altitude_tilt,
            self.azimuth_tilt,
            direct_horizontal_irradiance,
            np.float64(direct_normal_irradiance),
            direct_normal_irradiance_extra,
            apparent_zenith,
            sun_azimuth,
            airmass,
        )
        # calculate ground diffuse with specified albedo
        poa_ground_diffuse = pvlib.irradiance.get_ground_diffuse(
            self.altitude_tilt,
            global_horizontal_irradiance,
            albedo=albedo,
        )
        # calculate angle of incidence
        aoi = pvlib.irradiance.aoi(
            self.altitude_tilt,
            self.azimuth_tilt,
            apparent_zenith,
            sun_azimuth,
        )
        # calculate plane of array irradiance
        poa_irrad = pvlib.irradiance.poa_components(
            aoi,
            np.float64(direct_normal_irradiance),
            poa_sky_diffuse,
            poa_ground_diffuse,
        )

        if math.isnan(poa_irrad["poa_direct"]):
            self.incident_solar = 0
        else:
            self.incident_solar = (poa_irrad["poa_direct"]) * self.area

        solar_gains = (
            self.incident_solar
            * self.glass_solar_transmittance
            * self.nonperpendicular_reduction_factor
            * self.external_shading_vertical_reduction_factor
            * (1 - self.frame_area_fraction_reduction_factor)
        )
        return solar_gains

    def calc_direct_solar_factor(
        self,
        sun_altitude,
        sun_azimuth,
        apparent_zenith,
    ):
        """Calculates the cosine of the angle of incidence on the window.

        Commented equations, that provide a direct calculation, were derived in:

        Proportion of the radiation incident on the window (cos of the incident ray)
        ref:Quaschning, Volker, and Rolf Hanitsch. "Shade calculations in photovoltaic systems."
        ISES Solar World Conference, Harare. 1995.

        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        sun_altitude_rad = math.radians(sun_altitude)

        aoi = pvlib.irradiance.aoi(
            self.altitude_tilt,
            self.azimuth_tilt,
            apparent_zenith,
            sun_azimuth,
        )

        direct_factor = math.cos(aoi) / (math.sin(sun_altitude_rad))

        return direct_factor

    def calc_diffuse_solar_factor(
        self,
    ):
        """Calculates the proportion of diffuse radiation.

        Based on the RC_BuildingSimulator project @[rc_buildingsimulator-jayathissa] (** Check header)
        """
        # Proportion of incident light on the window surface
        return (1 + math.cos(self.altitude_tilt_rad)) / 2


class BuildingController(cp.Component):

    """BuildingController class.

    It calculates on base of the maximal Building
    Thermal Demand and the difference of the actual Building Tempreature
    to the Target/Minimal Building Tempreature how much the building is suppose
    to be heated up. This Output is called "RealBuildingHeatDemand".

    Parameters
    ----------
    sim_params : Simulator
        Simulator object used to carry the simulation using this class

    """

    # Inputs
    ReferenceMaxHeatBuildingDemand = "ReferenceMaxHeatBuildingDemand"
    ResidenceTemperature = "ResidenceTemperature"
    # Outputs
    RealHeatBuildingDemand = "RealHeatBuildingDemand"
    LevelOfUtilization = "LevelOfUtilization"

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: BuildingControllerConfig,
    ):
        """Constructs all the neccessary attributes of the Building Controller object."""
        super().__init__(
            name="BuildingController",
            my_simulation_parameters=my_simulation_parameters,
        )
        self.minimal_building_temperature_in_celsius = (
            config.minimal_building_temperature_in_celsius
        )
        self.stop_heating_building_temperature_in_celsius = (
            config.stop_heating_building_temperature_in_celsius
        )
        self.state = BuildingControllerState(
            temperature_building_target_in_celsius=config.minimal_building_temperature_in_celsius,
            level_of_utilization=0,
        )
        self.previous_state = self.state.clone()

        # =================================================================================================================================
        # Inputs and Output channels

        self.ref_max_thermal_build_demand_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ReferenceMaxHeatBuildingDemand,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
            True,
        )
        self.residence_temperature_channel: cp.ComponentInput = self.add_input(
            self.component_name,
            self.ResidenceTemperature,
            lt.LoadTypes.TEMPERATURE,
            lt.Units.CELSIUS,
            True,
        )
        self.real_heat_building_demand_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.RealHeatBuildingDemand,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
        )
        self.level_of_utilization_channel: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.LevelOfUtilization,
            lt.LoadTypes.ANY,
            lt.Units.PERCENT,
        )
        # =================================================================================================================================

    @staticmethod
    def get_default_config():
        """Gets a default configuration of the building controller."""
        config = BuildingControllerConfig(
            minimal_building_temperature_in_celsius=20,
            stop_heating_building_temperature_in_celsius=21,
        )
        return config

    def build(self):
        """Build load profile for entire simulation duration."""
        pass

    def write_to_report(
        self,
    ):
        """Writes a report."""
        pass

    def i_save_state(
        self,
    ):
        """Saves the current state."""
        self.previous_state = self.state.clone()

    def i_restore_state(
        self,
    ):
        """Restores previous state."""
        self.state = self.previous_state.clone()

    def i_doublecheck(
        self,
        timestep: int,
        stsv: cp.SingleTimeStepValues,
    ) -> None:
        """Doublechecks."""
        pass

    def i_prepare_simulation(
        self,
    ) -> None:
        """Prepares the simulation."""
        pass

    def i_simulate(
        self,
        timestep: int,
        stsv: cp.SingleTimeStepValues,
        force_convergence: bool,
    ) -> None:
        """Simulates the building controller."""
        building_temperature_in_celsius = stsv.get_input_value(
            self.residence_temperature_channel
        )
        minimal_building_temperature_in_celsius = (
            self.minimal_building_temperature_in_celsius
        )
        delta_temp_for_level_of_utilization = 0.4

        # Building is warm enough
        if building_temperature_in_celsius > minimal_building_temperature_in_celsius:
            level_of_utilization: float = 0
        # Building get heated up, when temperature is underneath target temperature
        elif (
            building_temperature_in_celsius
            < minimal_building_temperature_in_celsius
            - delta_temp_for_level_of_utilization
        ):
            level_of_utilization = 1
        else:
            level_of_utilization = (
                minimal_building_temperature_in_celsius
                - building_temperature_in_celsius
            )

        real_heat_building_demand_in_watt = (
            self.state.level_of_utilization
            * stsv.get_input_value(self.ref_max_thermal_build_demand_channel)
        )
        self.state.level_of_utilization = level_of_utilization
        stsv.set_output_value(self.level_of_utilization_channel, self.state.level_of_utilization)
        stsv.set_output_value(self.real_heat_building_demand_channel, real_heat_building_demand_in_watt)



""" Contains the base class for the charts. """
# clean
import os
import re


class Chart:  # noqa: too-few-public-methods

    """ Parent class for plots to be exported. """

    months_abbrev_uppercase = ['JAN', 'FEB', 'MAR', 'APR', 'MAY', 'JUN', 'JUL', 'AUG', 'SEP', 'OCT', 'NOV', 'DEZ']
    label_months_lowercase = ['January', 'February', 'March', 'April', 'May', 'June', 'July',
                              'August', 'September', 'October', 'November', 'December']

    def __init__(self, output, chart_type, units, directory_path, time_correction_factor, output2=None):
        """ Initializes the base class. """
        self.output = output
        self.type = chart_type
        if hasattr(units, "value"):
            self.units = units.value
            self.ylabel = units.value
        else:
            self.units = units
            self.ylabel = units
        self.time_correction_factor = time_correction_factor

        self.title: str = ""
        matches = re.finditer('.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$|#)', self.output)
        matches = [m.group(0) for m in matches]  # type: ignore
        pass_sign = False
        chart_property = ""
        chart_object = ""
        for single_match in matches:
            if pass_sign:
                chart_property = f"{chart_property} {single_match}"
            else:
                chart_object = f"{chart_object}{single_match}"

            if single_match.find("#"):  # type: ignore
                pass_sign = True

            if len(self.title) == 0:
                self.title = str(single_match)
            else:
                self.title = f"{self.title} {single_match}"
        self.directorypath = directory_path

        self.object_name = " "
        self.property = chart_property
        if output2 is not None:
            self.output2 = output2
            self.filename = f"{self.type.lower()}_{self.output.split(' # ', 2)[0]}_{self.output.split(' # ', 2)[1]}_double.png"
        else:
            self.filename = f"{self.type.lower()}_{self.output.split(' # ', 2)[0]}_{self.output.split(' # ', 2)[1]}.png"
        self.filepath = os.path.join(self.directorypath, self.filename)



""" Dynamic components are able to have an arbitrary number of inputs and outputs. """
# clean

from dataclasses import dataclass
from typing import List, Union, Any
from hisim import log

import hisim.loadtypes as lt
from hisim.component import Component, ComponentInput, SingleTimeStepValues, ComponentOutput


@dataclass
class DynamicConnectionInput:

    """ Class for describing a single component input. """

    source_component_class: str
    source_component_output: str
    source_load_type: lt.LoadTypes
    source_unit: lt.Units
    source_tags: list
    source_weight: int


@dataclass
class DynamicConnectionOutput:

    """ Describes a single component output for dynamic component. """

    source_component_class: str
    source_output_name: str
    source_tags: list
    source_weight: int
    source_load_type: lt.LoadTypes
    source_unit: lt.Units  # noqa


class DynamicComponent(Component):

    """ Class for components with a dynamic number of inputs and outputs. """

    def __init__(self, my_component_inputs, my_component_outputs, name, my_simulation_parameters):
        """ Initializes a dynamic component. """
        super().__init__(name=name, my_simulation_parameters=my_simulation_parameters)

        self.my_component_inputs = my_component_inputs
        self.my_component_outputs = my_component_outputs

    def add_component_input_and_connect(self,
                                        source_component_class: Component,
                                        source_component_output: str,
                                        source_load_type: lt.LoadTypes,
                                        source_unit: lt.Units,
                                        source_tags: List[Union[lt.ComponentType, lt.InandOutputType]],
                                        source_weight: int) -> None:
        """ Adds a component input and connects it at once. """
        # Label Input and generate variable
        num_inputs = len(self.inputs)
        label = f"Input{num_inputs}"
        vars(self)[label] = label
        log.trace("Added component input and connect: " + source_component_class.component_name + " - " + source_component_output)
        # Define Input as Component Input and add it to inputs
        myinput = ComponentInput(self.component_name, label, source_load_type, source_unit, True)
        self.inputs.append(myinput)
        myinput.src_object_name = source_component_class.component_name
        myinput.src_field_name = str(source_component_output)
        setattr(self, label, myinput)

        # Connect Input and define it as DynamicConnectionInput
        for output_var in source_component_class.outputs:
            if output_var.display_name == source_component_output:
                self.connect_input(label,
                                   source_component_class.component_name,
                                   output_var.field_name)
                self.my_component_inputs.append(DynamicConnectionInput(source_component_class=label,
                                                                       source_component_output=source_component_output,
                                                                       source_load_type=source_load_type,
                                                                       source_unit=source_unit,
                                                                       source_tags=source_tags,
                                                                       source_weight=source_weight))

    def add_component_inputs_and_connect(self,
                                         source_component_classes: List[Component],
                                         outputstring: str,
                                         source_load_type: lt.LoadTypes,
                                         source_unit: lt.Units,
                                         source_tags: List[Union[lt.ComponentType, lt.InandOutputType]],
                                         source_weight: int) -> None:
        """ Adds and connects inputs.

        Finds all outputs of listed components containing outputstring in outputname,
        adds inputs to dynamic component and connects the outputs.
        """

        # Label Input and generate variable
        num_inputs = len(self.inputs)

        # Connect Input and define it as DynamicConnectionInput
        for component in source_component_classes:
            for output_var in component.outputs:
                if outputstring in output_var.display_name:
                    source_component_output = output_var.display_name

                    label = f"Input{num_inputs}"
                    vars(self)[label] = label

                    # Define Input as Component Input and add it to inputs
                    myinput = ComponentInput(self.component_name, label, source_load_type, source_unit, True)
                    self.inputs.append(myinput)
                    myinput.src_object_name = component.component_name
                    myinput.src_field_name = str(source_component_output)
                    setattr(self, label, myinput)
                    num_inputs += 1
                    log.trace("Added component inputs and connect: " + myinput.src_object_name + " - " + myinput.src_field_name)
                    self.connect_input(label,
                                       component.component_name,
                                       output_var.field_name)
                    self.my_component_inputs.append(DynamicConnectionInput(source_component_class=label,
                                                                           source_component_output=source_component_output,
                                                                           source_load_type=source_load_type,
                                                                           source_unit=source_unit,
                                                                           source_tags=source_tags,
                                                                           source_weight=source_weight))

    def get_dynamic_input(self, stsv: SingleTimeStepValues,
                          tags: List[Union[lt.ComponentType, lt.InandOutputType]],
                          weight_counter: int) -> Any:
        """ Returns input value from first dynamic input with component type and weight. """
        inputvalue = None

        # check if component of component type is available
        for _, element in enumerate(self.my_component_inputs):  # loop over all inputs
            if all(tag in element.source_tags for tag in tags) and weight_counter == element.source_weight:
                inputvalue = stsv.get_input_value(getattr(self, element.source_component_class))
                break
        return inputvalue

    def get_dynamic_inputs(self, stsv: SingleTimeStepValues,
                           tags: List[Union[lt.ComponentType, lt.InandOutputType]]) -> List:
        """ Returns input values from all dynamic inputs with component type and weight. """
        inputvalues = []

        # check if component of component type is available
        for _, element in enumerate(self.my_component_inputs):  # loop over all inputs
            if all(tag in element.source_tags for tag in tags):
                inputvalues.append(stsv.get_input_value(getattr(self, element.source_component_class)))
            else:
                continue
        return inputvalues

    def set_dynamic_output(self, stsv: SingleTimeStepValues,
                           tags: List[Union[lt.ComponentType, lt.InandOutputType]],
                           weight_counter: int,
                           output_value: float) -> None:
        """ Sets all output values with given component type and weight. """

        # check if component of component type is available
        for _, element in enumerate(self.my_component_outputs):  # loop over all inputs
            if all(tag in element.source_tags for tag in tags) and weight_counter == element.source_weight:
                stsv.set_output_value(getattr(self, element.source_component_class), output_value)
            else:
                continue

    def add_component_output(self, source_output_name: str,
                             source_tags: list,
                             source_load_type: lt.LoadTypes,
                             source_unit: lt.Units,
                             source_weight: int) -> ComponentOutput:
        """ Adds an output channel to a component. """
        # Label Output and generate variable
        num_inputs = len(self.outputs)
        label = f"Output{num_inputs + 1}"
        vars(self)[label] = label

        # Define Output as Component Input and add it to inputs
        myoutput = ComponentOutput(object_name=self.component_name, field_name=source_output_name + label, load_type=source_load_type,
                                   unit=source_unit, sankey_flow_direction=True)
        self.outputs.append(myoutput)
        setattr(self, label, myoutput)

        # Define Output as DynamicConnectionInput
        self.my_component_outputs.append(DynamicConnectionOutput(source_component_class=label,
                                                                 source_output_name=source_output_name + label,
                                                                 source_tags=source_tags,
                                                                 source_load_type=source_load_type,
                                                                 source_unit=source_unit,
                                                                 source_weight=source_weight))
        return myoutput



from typing import Any, List
from hisim.component import Component
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim import utils

class CalculateOperation(cp.Component):
    operations_available = ["Sum", "Subtract", "Multiply", "Divide"]
    Output = "Output"

    def __init__(self, name: str, loadtype: lt.LoadTypes, unit: lt.Units, my_simulation_parameters: SimulationParameters )-> None:
        super().__init__(name=name, my_simulation_parameters=my_simulation_parameters)
        self.operations: List[str] = []
        self.loadtype = loadtype
        self.unit = unit
        self.output1: cp.ComponentOutput = self.add_output(self.component_name,
                                                           self.Output,
                                                           loadtype,
                                                           unit)

    def add_numbered_input(self) -> cp.ComponentInput:
        num_inputs = len(self.inputs)
        label = "Input{}".format(num_inputs + 1)
        vars(self)[label] = label
        myinput = cp.ComponentInput(self.component_name, label, self.loadtype, self.unit, True)
        self.inputs.append(myinput)
        return myinput

    def connect_arbitrary_input(self, src_object_name: str, src_field_name: str)-> None:
        next_input = self.add_numbered_input()
        next_input.src_object_name = src_object_name
        next_input.src_field_name = src_field_name

    def add_operation(self, operation: str) -> Any:
        num_operations = len(self.operations)
        num_inputs = len(self.inputs)
        if num_inputs == num_operations + 1:
            if operation in self.operations_available:
                self.operations.append(operation)
            else:
                raise Exception("Operation not implemented!")
        elif num_inputs >= num_operations + 1:
            raise Exception("Inputs connected without operation! {} operations are missing!".format(num_inputs - (num_operations + 1)))
        else:
            raise Exception("Inputs connected without operation! {} operations are missing!".format((num_operations + 1) - num_inputs))
        return operation

    def i_save_state(self)-> None:
        pass

    def i_restore_state(self)-> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues)-> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool)-> None:
        total:float = 0
        for index, input in enumerate(self.inputs):
            val1 = stsv.get_input_value(input)
            if index == 0:
                total = val1
            elif self.operations[index-1] == "Sum":
                    total = total + val1
            elif self.operations[index-1] == "Subtract":
                    total = total - val1
            elif self.operations[index-1] == "Multiply":
                    total = total * val1
            elif self.operations[index-1] == "Divide":
                    total = total / val1
            else:
                raise Exception("Operation invalid!")
        stsv.set_output_value(self.output1, total)

class ElectricityGrid(Component):
    operations_available = ["Sum", "Subtract"]
    ElectricityOutput = "ElectricityOutput"

    @utils.measure_execution_time
    def __init__(self, name: str, my_simulation_parameters: SimulationParameters , grid: Any=None, signal:Any=None) -> None:
        super().__init__(name="{}_{}".format("ElectricityGrid", name), my_simulation_parameters=my_simulation_parameters)
        self.signal=signal
        self.operations: List[str] = []
        self.loadtype = lt.LoadTypes.ELECTRICITY
        self.unit = lt.Units.WATT

        if grid is not None:
            self.connect_all(grid)

        self.electricity_outputC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                       self.ElectricityOutput,
                                                                       lt.LoadTypes.ELECTRICITY,
                                                                       lt.Units.WATT)
    def add_numbered_input(self) -> cp.ComponentInput:
        num_inputs = len(self.inputs)
        label = "Input{}".format(num_inputs + 1)
        vars(self)[label] = label
        myinput = cp.ComponentInput(self.component_name, label, self.loadtype, self.unit, True)
        self.inputs.append(myinput)
        return myinput

    def __add__(self, other_electricity_grid: Any) -> Any:
        return ElectricityGrid(name="{}Sum{}".format(self.component_name, other_electricity_grid.component_name), my_simulation_parameters=self.my_simulation_parameters,
                               grid=[self, "Sum", other_electricity_grid])

    def __sub__(self, other_electricity_grid: Any) -> Any:
        return ElectricityGrid(name="{}Subtract{}".format(self.component_name, other_electricity_grid.component_name),
                               grid=[self, "Subtract", other_electricity_grid], my_simulation_parameters=self.my_simulation_parameters)

    def write_to_report(self) -> List[str]:
        lines = []
        lines.append("Electricity grid")
        # todo: add more useful stuff here
        lines.append("tbd")
        return lines
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def connect_electricity_input(self, component: Component) -> None:
        if hasattr(component, 'ElectricityOutput') is False:
            raise Exception("Component does not contain electricity output.")
        next_input = self.add_numbered_input()
        next_input.src_object_name = component.component_name
        next_input.src_field_name = component.ElectricityOutput # type: ignore

    def add_operation(self, operation: str)-> Any:
        num_operations = len(self.operations)
        num_inputs = len(self.inputs)
        if num_inputs == num_operations + 1:
            if operation in self.operations_available:
                self.operations.append(operation)
            else:
                raise Exception("Operation not implemented!")
        elif num_inputs >= num_operations + 1:
            raise Exception("Inputs connected without operation! {} operations are missing!".format(num_inputs - (num_operations + 1)))
        else:
            raise Exception("Inputs connected without operation! {} operations are missing!".format((num_operations + 1) - num_inputs))
        return operation

    def connect_all(self, list_of_operations: Any) -> None:
        if isinstance(list_of_operations, list) is False:
            raise Exception("Input has to be a list!")
        elif len(list_of_operations) % 2 == 0:
            raise Exception("List of operations is incomplete!")

        for index, element in enumerate(list_of_operations):
            if index % 2 == 0:
                self.connect_electricity_input(element)
            else:
                self.add_operation(element)

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        total:float = 0
        for index, input in enumerate(self.inputs):
            val1 = stsv.get_input_value(input)
            if index == 0:
                total = val1
            elif self.operations[index-1] == "Sum":
                total = total + val1
            elif self.operations[index-1] == "Subtract":
                total = total - val1
            else:
                raise Exception("Operation invalid!")
        if self.signal == "Positive":
            if total <= 0:
                total = 0
        if self.signal == "Negative":
            if total >= 0:
                total = 0
        stsv.set_output_value(self.electricity_outputC, total)

class SumBuilderForTwoInputs(Component):
    SumInput1 = "Input 1"
    SumInput2 = "Input 2"
    SumOutput = "Sum"

    def __init__(self, name: str, loadtype: lt.LoadTypes, unit: lt.Units, my_simulation_parameters: SimulationParameters ) -> None:
        super().__init__(name=name, my_simulation_parameters=my_simulation_parameters)
        self.input1: cp.ComponentInput = self.add_input(self.component_name,
                                                        SumBuilderForTwoInputs.SumInput1,
                                                        loadtype,
                                                        unit,
                                                        True)
        self.input2: cp.ComponentInput = self.add_input(self.component_name,
                                                        SumBuilderForTwoInputs.SumInput2,
                                                        loadtype,
                                                        unit,
                                                        False)
        self.output1: cp.ComponentOutput = self.add_output(self.component_name,
                                                           SumBuilderForTwoInputs.SumOutput,
                                                           loadtype,
                                                           unit)

    def i_save_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_restore_state(self) -> None:
        pass
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        val1 = stsv.get_input_value(self.input1)
        val2 = stsv.get_input_value(self.input2)
        stsv.set_output_value(self.output1, val1+val2)
    
    def write_to_report(self) -> List[str]:
        lines =[]
        lines.append(f"Sumbuilder for two inputs: {self.component_name}")
        lines.append("Input 1: {}".format(self.input1.fullname))
        lines.append("Input 2: {}".format(self.input2.fullname))
        return lines

class SumBuilderForThreeInputs(Component):
    SumInput1 = "Input 1"
    SumInput2 = "Input 2"
    SumInput3 = "Input 3"
    SumOutput = "Sum"

    def __init__(self, name: str, loadtype: lt.LoadTypes, unit: lt.Units,my_simulation_parameters: SimulationParameters ) -> None:
        super().__init__(name=name, my_simulation_parameters=my_simulation_parameters)
        self.input1: cp.ComponentInput = self.add_input(self.component_name, SumBuilderForThreeInputs.SumInput1,
                                                        loadtype, unit, True)
        self.input2: cp.ComponentInput = self.add_input(self.component_name, SumBuilderForThreeInputs.SumInput2,
                                                        loadtype, unit, False)
        self.input3: cp.ComponentInput = self.add_input(self.component_name, SumBuilderForThreeInputs.SumInput3,
                                                        loadtype, unit, False)
        self.output1: cp.ComponentOutput = self.add_output(self.component_name, SumBuilderForThreeInputs.SumOutput,
                                                           loadtype, unit)

        self.state = 0
        self.previous_state = 0

    def i_save_state(self) -> None:
        pass
        #self.previous_state = copy.copy(self.state)

    def i_restore_state(self) -> None:
        pass
        #self.state = copy.copy(self.previous_state)

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues)  -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool)  -> None:
        val1 = stsv.get_input_value(self.input1)
        val2 = stsv.get_input_value(self.input2)
        val3 = stsv.get_input_value(self.input3)
        stsv.set_output_value(self.output1, val1+val2+val3)





# Generic/Built-in
import copy
import numpy as np

# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from dataclasses import dataclass
from dataclasses_json import dataclass_json
from typing import Any
__authors__ = "Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = ""
@dataclass_json
@dataclass
class HeatStorageConfig:
    V_SP_heating_water : float
    V_SP_warm_water : float
    temperature_of_warm_water_extratcion :float
    ambient_temperature :float
    T_sp_ww : float
    T_sp_hw :float

@dataclass_json
@dataclass
class HeatStorageControllerConfig:
    initial_temperature_building : float
    initial_temperature_heating_storage : float

class HeatStorageState:
    def __init__(self, T_sp_ww: float, T_sp_hw: float) -> None:
        self.T_sp_ww = T_sp_ww
        self.T_sp_hw = T_sp_hw
    def clone( self ) -> Any:
        return HeatStorageState( T_sp_ww = self.T_sp_ww,
                                 T_sp_hw = self.T_sp_hw )

class HeatStorage(Component):
    """
    This is a combined storage: buffer storage for heating, and hot water storage for hot water demand.
    It needs, hot water demand, heating demand, building temperature and a control signal choosing which of the two storages to heat as inputs.
    In addition it relies on Outside Temperature, and ThermalInputs of up to 5 heat sources.
    Based on this it evaluates the temperature in the storages based on the energy balance.
    """
    # Inputs
    ThermalDemandHeatingWater = "ThermalDemandHeatingWater"  # Heating Water to regulate room Temperature
    ThermalDemandWarmWater = "ThermalDemandHeating"  # Warmwater for showering, washing etc...
    ControlSignalChooseStorage = "ControlSignalChooseStorage"
    BuildingTemperature = "BuildingTemperature"

    OutsideTemperature = "OutsideTemperature"
    ThermalInputPower1 = "ThermalInputPower1"
    ThermalInputPower2 = "ThermalInputPower2"
    ThermalInputPower3 = "ThermalInputPower3"
    ThermalInputPower4 = "ThermalInputPower4"
    ThermalInputPower5 = "ThermalInputPower5"

    # Outputs
    WaterOutputTemperatureHeatingWater = "WaterOutputTemperatureHeatingWater"
    WaterOutputTemperatureWarmWater = "WaterOutputTemperatureWarmWater"
    WaterOutputStorageforHeaters = "WaterOutputStorageforHeaters"
    # StorageWarmWaterTemperature="StorageWarmWaterTemperature"
    StorageEnergyLoss = "StorageEnergyLoss"
    RealHeatForBuilding = "RealHeatForBuilding"

    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 config: HeatStorageConfig) -> None:
        super().__init__(name="HeatStorage", my_simulation_parameters=my_simulation_parameters)
        self.V_SP_heating_water = config.V_SP_heating_water
        self.V_SP_warm_water = config.V_SP_warm_water
        self.temperature_of_warm_water_extratcion = config.temperature_of_warm_water_extratcion
        self.ambient_temperature = config.ambient_temperature
        self.cw = 4812

        self.state = HeatStorageState(config.T_sp_ww, config.T_sp_hw)
        self.previous_state = self.state.clone( )

        self.thermal_demand_heating_water: ComponentInput = self.add_input(self.component_name,
                                                                           self.ThermalDemandHeatingWater,
                                                                           lt.LoadTypes.WARM_WATER,
                                                                           lt.Units.WATT,
                                                                           False)

        self.thermal_demand_warm_water: ComponentInput = self.add_input(self.component_name,
                                                                        self.ThermalDemandWarmWater,
                                                                        lt.LoadTypes.WARM_WATER,
                                                                        lt.Units.WATT,
                                                                        False)
        self.control_signal_choose_storage: cp.ComponentInput = self.add_input(self.component_name,
                                                                               self.ControlSignalChooseStorage,
                                                                               lt.LoadTypes.ANY,
                                                                               lt.Units.ANY,
                                                                               False)
        self.building_temperature: cp.ComponentInput = self.add_input(self.component_name,
                                                                      self.BuildingTemperature,
                                                                      lt.LoadTypes.TEMPERATURE,
                                                                      lt.Units.CELSIUS,
                                                                      False)
        self.thermal_input_power1: ComponentInput = self.add_input(self.component_name,
                                                                   self.ThermalInputPower1,
                                                                   lt.LoadTypes.HEATING,
                                                                   lt.Units.WATT,
                                                                   False)
        self.thermal_input_power2: ComponentInput = self.add_input(self.component_name,
                                                                   self.ThermalInputPower2,
                                                                   lt.LoadTypes.HEATING,
                                                                   lt.Units.WATT,
                                                                   False)
        self.thermal_input_power3: ComponentInput = self.add_input(self.component_name,
                                                                   self.ThermalInputPower3,
                                                                   lt.LoadTypes.HEATING,
                                                                   lt.Units.WATT,
                                                                   False)
        self.thermal_input_power4: ComponentInput = self.add_input(self.component_name,
                                                                   self.ThermalInputPower4,
                                                                   lt.LoadTypes.HEATING,
                                                                   lt.Units.WATT,
                                                                   False)
        self.thermal_input_power5: ComponentInput = self.add_input(self.component_name,
                                                                   self.ThermalInputPower5,
                                                                   lt.LoadTypes.HEATING,
                                                                   lt.Units.WATT,
                                                                   False)

        self.T_sp_C_hw: ComponentOutput = self.add_output(self.component_name,
                                                          self.WaterOutputTemperatureHeatingWater,
                                                          lt.LoadTypes.TEMPERATURE,
                                                          lt.Units.CELSIUS)
        self.T_sp_C_ww: ComponentOutput = self.add_output(self.component_name,
                                                          self.WaterOutputTemperatureWarmWater,
                                                          lt.LoadTypes.TEMPERATURE,
                                                          lt.Units.CELSIUS)
        self.UA_SP_C: ComponentOutput = self.add_output(self.component_name,
                                                        self.StorageEnergyLoss,
                                                        lt.LoadTypes.ANY,
                                                        lt.Units.WATT)
        self.T_sp_C: ComponentOutput = self.add_output(self.component_name,
                                                       self.WaterOutputStorageforHeaters,
                                                       lt.LoadTypes.TEMPERATURE,
                                                       lt.Units.CELSIUS)
        self.real_heat_for_building: ComponentOutput = self.add_output(self.component_name,
                                                                       self.RealHeatForBuilding,
                                                                       lt.LoadTypes.HEATING,
                                                                       lt.Units.WATT)
    @staticmethod
    def get_default_config():
        config=HeatStorageConfig(
                V_SP_heating_water = 1000,
                V_SP_warm_water = 100,
                temperature_of_warm_water_extratcion = 32,
                ambient_temperature = 15,
                T_sp_ww=40,
                T_sp_hw=40)
        return config
    def write_to_report(self) -> None:
        pass
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_save_state(self) -> None:
        self.previous_state = self.state.clone( )

    def i_restore_state(self) -> None:
        self.state = self.previous_state.clone( )

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def adding_all_possible_mass_flows(self, stsv: cp.SingleTimeStepValues, c_w: float) -> Any:
        production:float = 0
        # function to add all possible mass flows

        if self.thermal_input_power1.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power1) + production

        if self.thermal_input_power2.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power2) + production

        if self.thermal_input_power3.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power3) + production

        if self.thermal_input_power4.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power4) + production

        if self.thermal_input_power5.source_output is not None:
            production = stsv.get_input_value(self.thermal_input_power5) + production

        return production

    def calculate_new_storage_temperature(self, seconds_per_timestep: int, T_sp: float, production: float, last: float,
                                          c_w: float, V_SP: float) -> Any:

        T_ext_SP = self.ambient_temperature

        m_SP_h = V_SP * 0.99  # Vereinfachung
        UA_SP = 0.0038 * V_SP + 0.85  # Heatloss Storage
        dt = seconds_per_timestep

        # calcutae new Storage Temp.
        T_SP = T_sp + (1 / (m_SP_h * c_w)) * (production - last - UA_SP * (T_sp - T_ext_SP)) * dt
        # T_SP = T_sp + (dt/(m_SP_h*c_w))*(P_h_HS*(T_sp-T_ext_SP) - last*(T_sp-T_ext_SP) - UA_SP*(T_sp-T_ext_SP))
        # Correction Calculation
        # T_sp_k = (T_sp+T_SP)/2
        # T_vl = T_sp_k+2.5

        # calcutae new Storage Temp.
        # T_SP = T_sp + (1/(m_SP_h*c_w))*( production- last - UA_SP*(T_sp_k-T_ext_SP))*dt

        return T_SP, UA_SP

    # def regarding_heating_water_storage (self, T_sp: int):

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:

        T_sp_var_ww = self.state.T_sp_ww  # Start-Temp-Storage
        T_sp_var_hw = self.state.T_sp_hw  # Start-Temp-Storage

        last_var_ww = stsv.get_input_value(self.thermal_demand_warm_water)
        last_var_hw = stsv.get_input_value(self.thermal_demand_heating_water)

        result_ww = [T_sp_var_ww, 0]
        result_hw = [T_sp_var_hw, 0]
        T_sp_C = (T_sp_var_ww + T_sp_var_hw) / 2

        if stsv.get_input_value(self.control_signal_choose_storage) == 1:  # choose to heat up warm water storage
            production_var = self.adding_all_possible_mass_flows(stsv, c_w=self.cw)
            result_ww = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep, T_sp=T_sp_var_ww,
                production=production_var, last=last_var_ww, c_w=self.cw, V_SP=self.V_SP_warm_water)
            T_sp_C = result_ww[0]
            production_var = 0
            result_hw = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep, T_sp=T_sp_var_hw,
                production=production_var, last=last_var_hw, c_w=self.cw, V_SP=self.V_SP_heating_water)

        elif stsv.get_input_value(self.control_signal_choose_storage) == 2:  # choose to heat up heating water storage
            production_var = self.adding_all_possible_mass_flows(stsv, c_w=self.cw)
            result_hw = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep, T_sp=T_sp_var_hw,
                production=production_var, last=last_var_hw, c_w=self.cw, V_SP=self.V_SP_heating_water)

            T_sp_C = result_hw[0]
            production_var = 0
            result_ww = self.calculate_new_storage_temperature(
                seconds_per_timestep=self.my_simulation_parameters.seconds_per_timestep, T_sp=T_sp_var_ww,
                production=production_var, last=last_var_ww, c_w=self.cw, V_SP=self.V_SP_warm_water)

        self.state.T_sp_ww = result_ww[0]
        self.state.T_sp_hw = result_hw[0]
        stsv.set_output_value(self.T_sp_C_ww, self.state.T_sp_ww)
        stsv.set_output_value(self.T_sp_C_hw, self.state.T_sp_hw)
        stsv.set_output_value(self.T_sp_C, T_sp_C)
        stsv.set_output_value(self.UA_SP_C, result_ww[1] + result_hw[1])
        stsv.set_output_value(self.real_heat_for_building, last_var_hw)

        # Output Massenstrom von Wasser entspricht dem Input Massenstrom. Nur Temperatur hat sich geändert. Wie ist das zu behandelN?



class HeatStorageController(cp.Component):
    """
    HeatStorageController class calculates on base of the maximal Building
    Thermal Demand and the TemperatureHeatingStorage and Building Tempreature
    the real thermal demand for the Heating Storage.
    This Output is called "RealThermalDemandHeatingWater".

    Parameters
    ----------
    sim_params : Simulator
        Simulator object used to carry the simulation using this class
    """

    # Inputs
    ReferenceMaxHeatBuildingDemand = "ReferenceMaxHeatBuildingDemand"
    TemperatureHeatingStorage = "TemperatureHeatingStorage"
    BuildingTemperature = "BuildingTemperature"
    RealHeatBuildingDemand= "RealHeatBuildingDemand"
    # Outputs
    RealThermalDemandHeatingWater = "RealThermalDemandHeatingWater"


    def __init__(self,
                 my_simulation_parameters: SimulationParameters,
                 config: HeatStorageControllerConfig
                 ) -> None:
        super().__init__(name="HeatStorageController", my_simulation_parameters=my_simulation_parameters)
        self.initial_temperature_heating_storage = config.initial_temperature_heating_storage
        self.initial_temperature_building = config.initial_temperature_building
        # ===================================================================================================================
        # Inputs
        self.ref_max_thermal_build_demand: ComponentInput = self.add_input(self.component_name,
                                                                           self.ReferenceMaxHeatBuildingDemand,
                                                                           lt.LoadTypes.HEATING,
                                                                           lt.Units.WATT,
                                                                           False)
        self.heating_storage_temperature: ComponentInput = self.add_input(self.component_name,
                                                                          self.TemperatureHeatingStorage,
                                                                          lt.LoadTypes.TEMPERATURE,
                                                                          lt.Units.CELSIUS,
                                                                          False)
        self.building_temperature: ComponentInput = self.add_input(self.component_name,
                                                                   self.BuildingTemperature,
                                                                   lt.LoadTypes.TEMPERATURE,
                                                                   lt.Units.CELSIUS,
                                                                   False)
        self.real_thermal_demand_building = self.add_input(self.component_name,
                                                           self.RealHeatBuildingDemand,
                                                           lt.LoadTypes.HEATING,
                                                           lt.Units.WATT,
                                                           False)
        # Outputs
        self.real_thermal_demand_heating_water: ComponentOutput = self.add_output(self.component_name,
                                                                                  self.RealThermalDemandHeatingWater,
                                                                                  lt.LoadTypes.HEATING,
                                                                                  lt.Units.WATT)
    @staticmethod
    def get_default_config() -> HeatStorageControllerConfig:
        config=HeatStorageControllerConfig(
                initial_temperature_building = 20,
                initial_temperature_heating_storage = 35)
        return config
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def build(self):
        pass

    def write_to_report(self) -> None:
        pass

    def i_save_state(self) -> None:
        pass

    def i_restore_state(self) -> None:
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        T_sp_var_hw = stsv.get_input_value(self.heating_storage_temperature)  # Start-Temp-Storage
        last_var_hw = stsv.get_input_value(self.real_thermal_demand_building)
        max_mass_flow_heat_storage = stsv.get_input_value(self.ref_max_thermal_build_demand) / (
                    4.1851 * 1000 * (self.initial_temperature_heating_storage - self.initial_temperature_building))

        max_last_var_hw = max_mass_flow_heat_storage * 4.185 * 1000 * (
                    T_sp_var_hw - stsv.get_input_value(self.building_temperature))

        if max_last_var_hw < last_var_hw:
            last_var_hw = max_last_var_hw


        stsv.set_output_value(self.real_thermal_demand_heating_water, last_var_hw)



""" Generates Class for the results of the modular_example preprocessing."""

from dataclasses import dataclass
from dataclasses_json import dataclass_json
import hisim.loadtypes as lt


@dataclass_json
@dataclass()
class ModularHouseholdResults:

    """Class for Results of ModularHousehold."""

    investment_cost: float
    co2_cost: float
    injection: float
    autarky_rate: float
    self_consumption_rate: float
    terminationflag: lt.Termination = lt.Termination.SUCCESSFUL  # add enum in loadtypes



"""The ``template`` module.

It serves as a template for creating new component modules.
It shows with a simplified example which steps are necessary to create a new component.
Additionally it contains examples for doc strings according to the sphinx format.

"""

# clean

# Import packages from standard library or the environment e.g. pandas, numpy etc.
from copy import deepcopy
from dataclasses import dataclass
from dataclasses_json import dataclass_json

# Import modules from HiSim
from hisim.component import (
    Component,
    ComponentInput,
    ComponentOutput,
    SingleTimeStepValues,
)
from hisim import loadtypes
from hisim.simulationparameters import SimulationParameters
from hisim.component import ConfigBase

__authors__ = "Tjarko Tjaden, Kai Rösken"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class ComponentNameConfig(ConfigBase):

    """Configuration of the ComponentName."""

    @classmethod
    def get_main_classname(cls):
        """Returns the full class name of the base class."""
        return ComponentName.get_full_classname()

    # parameter_string: str
    # my_simulation_parameters: SimulationParameters
    name: str
    loadtype: loadtypes.LoadTypes
    unit: loadtypes.Units

    @classmethod
    def get_default_template_component(cls):
        """Gets a default ComponentName."""
        return ComponentNameConfig(
            name="ComponentName default",
            loadtype=loadtypes.LoadTypes.ELECTRICITY,
            unit=loadtypes.Units.WATT,
        )


class ComponentName(Component):

    """Some instructions to document a class.

    First write a summary of the class that is as accurate as possible.
    It should contain information about what functionalities the class has
    and what its purpose is in the HiSim project.

    This is an example class. It can be used as a template for creating new
    component modules. Its functionalities serve no further purpose for HiSim.

    Parameters
    ----------
    component_name : str
        Passed to initialize :py:class:`~hisim.component.Component`.

    loadtype : LoadType
        A :py:class:`~hisim.loadtypes.LoadTypes` object that represents
        the type of the loaded data.

    unit: LoadTypes.Units
        A :py:class:`~hisim.loadtypes.Units` object that represents
        the unit of the loaded data.

    """

    # Inputs
    InputFromOtherComponent = "InputFromState"

    # Outputs
    OutputWithState = "OutputWithState"
    OutputWithoutState = "OutputWithoutState"

    # def __init__(self, component_name: str, my_simulation_parameters: SimulationParameters) -> None:
    #     """Constructs all the neccessary attributes for the ExampleStorage object."""
    #     super().__init__(name=component_name, my_simulation_parameters=my_simulation_parameters)

    #     # If a component requires states, this can be implemented here.
    #     self.state = ComponentNameState()
    #     self.previous_state = deepcopy(self.state)

    #     self.input_from_other_component: ComponentInput = self.add_input(
    #         object_name=self.component_name,
    #         field_name=self.InputFromOtherComponent,
    #         load_type=LoadTypes.ELECTRICITY,
    #         unit=Units.WATT,
    #         mandatory=True,
    #     )

    #     self.output_with_state: ComponentOutput = self.add_output(
    #         object_name=self.component_name,
    #         field_name=self.OutputWithState,
    #         load_type=LoadTypes.ELECTRICITY,
    #         unit=Units.WATT_HOUR,
    #     )

    #     self.output_without_state: ComponentOutput = self.add_output(
    #         object_name=self.component_name,
    #         field_name=self.OutputWithoutState,
    #         load_type=LoadTypes.ELECTRICITY,
    #         unit=Units.WATT,
    #     )
    #     self.factor = 1.0

    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: ComponentNameConfig,
    ) -> None:
        """Constructs all the neccessary attributes."""
        self.componentnameconfig = config
        super().__init__(
            self.componentnameconfig.name,
            my_simulation_parameters=my_simulation_parameters,
        )

        # If a component requires states, this can be implemented here.
        self.state = ComponentNameState()
        self.previous_state = deepcopy(self.state)
        # Initialized variables
        self.factor = 1.0

        self.input_from_other_component: ComponentInput = self.add_input(
            object_name=self.componentnameconfig.name,
            field_name=self.InputFromOtherComponent,
            load_type=loadtypes.LoadTypes.ELECTRICITY,
            unit=loadtypes.Units.WATT,
            mandatory=True,
        )

        self.output_with_state: ComponentOutput = self.add_output(
            object_name=self.componentnameconfig.name,
            field_name=self.OutputWithState,
            load_type=loadtypes.LoadTypes.ELECTRICITY,
            unit=loadtypes.Units.WATT_HOUR,
        )

        self.output_without_state: ComponentOutput = self.add_output(
            object_name=self.componentnameconfig.name,
            field_name=self.OutputWithoutState,
            load_type=loadtypes.LoadTypes.ELECTRICITY,
            unit=loadtypes.Units.WATT_HOUR,
        )

    def i_save_state(self) -> None:
        """Saves the current state."""
        self.previous_state = deepcopy(self.state)

    def i_restore_state(self) -> None:
        """Restores previous state."""
        self.state = deepcopy(self.previous_state)

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        """Doublechecks."""
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        """Simulates the component."""
        # define local variables
        input_1 = stsv.get_input_value(self.input_from_other_component)
        input_2 = self.state.output_with_state

        # do your calculations
        output_1 = input_2 + input_1 * self.my_simulation_parameters.seconds_per_timestep
        output_2 = input_1 + self.factor

        # write values for output time series
        stsv.set_output_value(self.output_with_state, output_1)
        stsv.set_output_value(self.output_without_state, output_2)

        # write values to state
        self.state.output_with_state = output_1


@dataclass
class ComponentNameState:

    """The data class saves the state of the simulation results.

    Parameters
    ----------
    output_with_state : int
    Stores the state of the output_with_state value from
    :py:class:`~hisim.component.ComponentName`.

    """

    output_with_state: float = 0



"""Simple Car (LPG connected) and charging station (if it is electric)."""

# -*- coding: utf-8 -*-
from typing import List, Any
from os import listdir, path
from collections import Counter
from dataclasses import dataclass
import datetime as dt
import json
from dataclasses_json import dataclass_json

import pandas as pd
import numpy as np

from hisim import component as cp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim import utils

__authors__ = "Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Johanna Ganglbauer"
__email__ = "johanna.ganglbauer@4wardenergy.at"
__status__ = "development"


@dataclass_json
@dataclass
class CarConfig:

    """ Definition of configuration of Car. """

    name: str
    source_weight: int
    fuel: lt.LoadTypes
    consumption_per_km: float

    def __init__(self, name: str, source_weight: int, fuel: lt.LoadTypes, consumption_per_km: float):
        """ Initialization of car configuration. """

        self.name = name
        self.source_weight = source_weight
        self.fuel = fuel
        self.consumption_per_km = consumption_per_km  # consumption in kWh/km or l/km

    @staticmethod
    def get_default_diesel_config() -> Any:
        """ Defines default configuration for diesel vehicle. """
        config = CarConfig(name='Car', source_weight=1, fuel=lt.LoadTypes.DIESEL, consumption_per_km=0.06)
        return config

    @staticmethod
    def get_default_ev_config() -> Any:
        """ Defines default configuration for electric vehicle. """
        config = CarConfig(name='Car', source_weight=1, fuel=lt.LoadTypes.ELECTRICITY, consumption_per_km=0.15)
        return config


class Car(cp.Component):

    """ Simulates car with constant consumption. Car usage (driven kilometers and state) orginate from LPG. """

    # Outputs
    FuelConsumption = "FuelConsumption"
    ElectricityOutput = "ElectricityOutput"
    CarLocation = "CarLocation"

    def __init__(self, my_simulation_parameters: SimulationParameters, config: CarConfig,
                 occupancy_config: Any) -> None:
        """ Initializes Car. """
        super().__init__(name=config.name + '_w' + str(config.source_weight),
                         my_simulation_parameters=my_simulation_parameters)
        self.build(config=config, occupancy_config=occupancy_config)

        if self.fuel == lt.LoadTypes.ELECTRICITY:
            self.electricity_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name, field_name=self.ElectricityOutput, load_type=lt.LoadTypes.ELECTRICITY,
                unit=lt.Units.WATT)
            self.car_location_output: cp.ComponentOutput = self.add_output(
                object_name=self.component_name, field_name=self.CarLocation, load_type=lt.LoadTypes.ANY,
                unit=lt.Units.ANY)
        elif self.fuel == lt.LoadTypes.DIESEL:
            self.fuel_consumption: cp.ComponentOutput = self.add_output(
                object_name=self.component_name, field_name=self.FuelConsumption, load_type=lt.LoadTypes.DIESEL,
                unit=lt.Units.LITER)

    def i_save_state(self) -> None:
        """ Saves actual state. """
        pass

    def i_restore_state(self) -> None:
        """ Restores previous state. """
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """ Checks statements. """
        pass

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        """ Returns consumption and location of car in each timestep. """

        if self.fuel == lt.LoadTypes.ELECTRICITY:
            watt_used = self.meters_driven[timestep] * self.consumption_per_km * \
                (3600 / self.my_simulation_parameters.seconds_per_timestep)  # conversion Wh to W
            stsv.set_output_value(self.electricity_output, watt_used)
            stsv.set_output_value(self.car_location_output, self.car_location[timestep])

        # if not already running: check if activation makes sense
        elif self.fuel == lt.LoadTypes.DIESEL:
            liters_used = self.meters_driven[timestep] * self.consumption_per_km * 1e-3  # conversion meter to kilometer
            stsv.set_output_value(self.fuel_consumption, liters_used)

    def build(self, config: CarConfig, occupancy_config: Any) -> None:
        """ Loads necesary data and saves config to class. """
        self.name = config.name
        self.source_weight = config.source_weight
        self.fuel = config.fuel
        self.consumption_per_km = config.consumption_per_km
        self.car_location = []
        self.meters_driven = []

        location_translator = {"School": 0, "Event Location": 0, "Shopping": 0, None: 0,
                               "Home": 1, "Workplace": 2}

        # check if caching is possible
        file_exists, cache_filepath = utils.get_cache_file(component_key=self.component_name, parameter_class=occupancy_config,
                                                           my_simulation_parameters=self.my_simulation_parameters)
        if file_exists:
            # load from cache
            dataframe = pd.read_csv(cache_filepath, sep=',', decimal='.', encoding="cp1252")
            self.car_location = dataframe['car_location'].tolist()
            self.meters_driven = dataframe['meters_driven'].tolist()
        else:
            # load car data from LPG output
            filepaths = listdir(utils.HISIMPATH["utsp_results"])
            filepath_location = [elem for elem in filepaths if "CarLocation." + self.name in elem][0]
            filepath_meters_driven = [elem for elem in filepaths if "DrivingDistance." + self.name in elem][0]
            with open(path.join(utils.HISIMPATH["utsp_results"], filepath_location), encoding="utf-8") as json_file:
                car_location = json.load(json_file)
            with open(path.join(utils.HISIMPATH["utsp_results"], filepath_meters_driven), encoding="utf-8") as json_file:
                meters_driven = json.load(json_file)

            # compare time resolution of LPG to time resolution of hisim
            time_resolution_original = dt.datetime.strptime(car_location["TimeResolution"], "%H:%M:%S")
            seconds_per_timestep_original = time_resolution_original.hour * 3600 \
                + time_resolution_original.minute * 60 + time_resolution_original.second
            steps_ratio = int(self.my_simulation_parameters.seconds_per_timestep / seconds_per_timestep_original)

            # extract values for location and distance of car
            car_location = car_location["Values"]
            meters_driven = meters_driven["Values"]

            # translate car location to integers (according to location_translator)
            car_location = [location_translator[elem] for elem in car_location]

            # sum / extract most common value from data to match hisim time resolution
            for i in range(int(len(meters_driven) / steps_ratio)):
                self.meters_driven.append(sum(meters_driven[i * steps_ratio: (i + 1) * steps_ratio]))  # sum
                location_list = car_location[i * steps_ratio: (i + 1) * steps_ratio]  # extract list
                occurence_count: Counter = Counter(location_list)  # extract most common
                self.car_location.append(occurence_count.most_common(1)[0][0])

            # save data in cache
            data = np.transpose([self.car_location,
                                 self.meters_driven])
            database = pd.DataFrame(data, columns=['car_location',
                                                   'meters_driven'])

            database.to_csv(cache_filepath)
            del data
            del database

    def write_to_report(self) -> List[str]:
        """ Writes Car values to report. """
        lines = []
        lines.append("LPG configured" + self.fuel.value + " " + self.component_name)
        return lines



# -*- coding: utf-8 -*-

# Generic/Built-in
import numpy as np
from typing import Optional, Any

# Owned
import hisim.utils as utils
from hisim import component as cp
from hisim.loadtypes import LoadTypes, Units
from hisim.simulationparameters import SimulationParameters
from hisim.components import controller_l1_generic_runtime
from hisim.components.building import Building
from hisim.components import generic_hot_water_storage_modular
from hisim import log

from dataclasses import dataclass
from dataclasses_json import dataclass_json

__authors__ = "edited Johanna Ganglbauer"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class L2HeatSmartConfig:
    """
    L2 Config
    """
    name: str
    source_weight: int
    T_min_heating: float
    T_max_heating: float
    T_tolerance: float
    P_threshold: float
    cooling_considered: bool
    T_min_cooling: Optional[float]
    T_max_cooling: Optional[float]
    heating_season_begin: Optional[int]
    heating_season_end: Optional[int]

    def __init__( self,
                  name: str,
                  source_weight: int,
                  T_min_heating: float,
                  T_max_heating: float,
                  T_tolerance: float,
                  P_threshold: float,
                  cooling_considered: bool,
                  T_min_cooling: Optional[float],
                  T_max_cooling: Optional[float],
                  heating_season_begin: Optional[int],
                  heating_season_end: Optional[int] ):
        self.name = name
        self.source_weight = source_weight
        self.T_min_heating = T_min_heating
        self.T_max_heating = T_max_heating
        self.T_tolerance = T_tolerance
        self.P_threshold = P_threshold
        self.cooling_considered = cooling_considered
        self.T_min_cooling = T_min_cooling
        self.T_max_cooling = T_max_cooling
        self.heating_season_begin = heating_season_begin
        self.heating_season_end = heating_season_end

class L2HeatSmartControllerState:
    """
    This data class saves the state of the heat pump.
    """

    def __init__( self, timestep_actual : int = -1, state : int = 0, compulsory : int = 0, count : int = 0 ):
        self.timestep_actual = timestep_actual
        self.state = state
        self.compulsory = compulsory
        self.count = count
        
    def clone( self ):
        return L2HeatSmartControllerState(timestep_actual = self.timestep_actual, state = self.state, compulsory = self.compulsory, count = self.count)
    
    def is_first_iteration( self, timestep ):
        if self.timestep_actual + 1 == timestep:
            self.timestep_actual += 1
            self.compulsory = 0
            self.count = 0
            return True
        else:
            return False
        
    def is_compulsory( self ):
        if self.count <= 1:
            self.compulsory = 0
        else:
            self.compulsory = 1
        
    
    def activate( self ):
        self.state = 1
        self.compulsory = 1
        self.count += 1
        
    def deactivate( self ):
        self.state = 0
        self.compulsory = 1
        self.count += 1

class L2HeatSmartController(cp.Component):
    
    """ L2 heat pump controller. Processes signals ensuring comfort temperature of building.
    Gets available surplus electricity and the temperature of the storage or building to control as input,
    and outputs control signal 0/1 for turn off/switch on based on comfort temperature limits and available electricity.
    It optionally has different modes for cooling and heating selected by the time of the year.

    Parameters
    --------------
    source_weight : int, optional
        Weight of component, relevant if there is more than one component of same type, defines hierachy in control. The default is 1.
    T_min_heating: float, optional
        Minimum comfortable temperature for residents during heating period, in °C. The default is 19 °C.
    T_max_heating: float, optional
        Maximum comfortable temperature for residents during heating period, in °C. The default is 23 °C.
    T_tolerance : float, optional
        Temperature difference the building may go below or exceed the comfort temperature band with, because of recommendations from L3. The default is 1 °C.
    P_threshold : float, optional
        Estimated power to drive heat source. The defauls is 1500 W.
    T_min_cooling: float, optional
        Minimum comfortable temperature for residents during cooling period, in °C. The default is 23 °C.
    T_max_cooling: float, optional
        Maximum comfortable temperature for residents during cooling period, in °C. The default is 26 °C.
    heating_season_begin : int, optional
        Day( julian day, number of day in year ), when heating season starts - and cooling season ends. The default is 270.
    heating_season_end : int, optional
        Day( julian day, number of day in year ), when heating season ends - and cooling season starts. The default is 150.
    """
    # Inputs
    ReferenceTemperature = "ReferenceTemperature"
    ElectricityTarget = "ElectricityTarget"
    L1RunTimeSignal = "L1RunTimeSignal"

    # Outputs
    l2_DeviceSignal = "l2_DeviceSignal"

    # #Forecasts
    # HeatPumpLoadForecast = "HeatPumpLoadForecast"

    # Similar components to connect to:
    # 1. Building
    # 2. HeatPump
    
    @utils.measure_execution_time
    def __init__(self, my_simulation_parameters : SimulationParameters, config: L2HeatSmartConfig) -> None:
        if not config.__class__.__name__ == L2HeatSmartConfig.__name__:
            raise ValueError("Wrong config class: " + config.__class__.__name__)
        super().__init__(name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters)
        self.build(config)
        
        #Component Outputs
        self.l2_DeviceSignalC: cp.ComponentOutput = self.add_output(self.component_name,
                                                                    self.l2_DeviceSignal,
                                                                    LoadTypes.ON_OFF,
                                                                    Units.BINARY)


        #Component Inputs
        self.ReferenceTemperatureC: cp.ComponentInput = self.add_input(self.component_name,
                                                                       self.ReferenceTemperature,
                                                                       LoadTypes.TEMPERATURE,
                                                                       Units.CELSIUS,
                                                                       mandatory=True)
        self.L1RunTimeSignalC: cp.ComponentInput = self.add_input(
            self.component_name, self.L1RunTimeSignal, LoadTypes.ANY, Units.ANY, True)
        
        self.add_default_connections( Building, self.get_building_default_connections( ) )
        self.add_default_connections( generic_hot_water_storage_modular.HotWaterStorage, self.get_boiler_default_connections( ) )
        self.add_default_connections(controller_l1_generic_runtime.L1GenericRuntimeController, self.get_l1_default_connections())
        
        self.ElectricityTargetC: cp.ComponentInput = self.add_input( self.component_name,
                                                                      self.ElectricityTarget,
                                                                      LoadTypes.ELECTRICITY,
                                                                      Units.WATT,
                                                                      mandatory = True)

    def get_building_default_connections( self ):
        log.information("setting building default connections in L2 Controller")
        connections = [ ]
        building_classname = Building.get_classname( )
        connections.append(cp.ComponentConnection(L2HeatSmartController.ReferenceTemperature, building_classname, Building.TemperatureMean))
        return connections
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def get_boiler_default_connections( self ):
        log.information("setting boiler default connections in L2 Controller")
        connections = [ ]
        boiler_classname = generic_hot_water_storage_modular.HotWaterStorage.get_classname( )
        connections.append(cp.ComponentConnection(L2HeatSmartController.ReferenceTemperature, boiler_classname, generic_hot_water_storage_modular.HotWaterStorage.TemperatureMean))
        return connections
    def get_l1_default_connections( self ):
        log.information("setting L1 default connections in L2 Controller")
        connections = []
        l1_classname = controller_l1_generic_runtime.L1GenericRuntimeController.get_classname()
        connections.append(cp.ComponentConnection(L2HeatSmartController.l1_RunTimeSignal, l1_classname, controller_l1_generic_runtime.L1GenericRuntimeController.l1_RunTimeSignal))
        return connections
    
    @staticmethod
    def get_default_config_heating():
        config = L2HeatSmartConfig(name ='L2HeatingTemperatureController',
                                   source_weight =  1,
                                   T_min_heating = 20.0,
                                   T_max_heating = 22.0,
                                   T_tolerance = 1.0,
                                   P_threshold = 1500,
                                   cooling_considered = False,
                                   T_min_cooling = 23.0,
                                   T_max_cooling = 25.0,
                                   heating_season_begin = 270,
                                   heating_season_end = 150)
        return config
    
    @staticmethod
    def get_default_config_buffer_heating():
        config = L2HeatSmartConfig(name ='L2BufferTemperatureController',
                                   source_weight =  1,
                                   T_min_heating = 40.0,
                                   T_max_heating = 60.0,
                                   T_tolerance = 10.0,
                                   P_threshold = 1500,
                                   cooling_considered = False,
                                   T_min_cooling = 5.0,
                                   T_max_cooling = 15.0,
                                   heating_season_begin = 270,
                                   heating_season_end = 150)
        return config
    
    @staticmethod
    def get_default_config_waterheating():
        config = L2HeatSmartConfig(name ='L2DHWTemperatureController',
                                   source_weight =  1,
                                   T_min_heating = 50.0,
                                   T_max_heating = 80.0,
                                   T_tolerance = 5.0,
                                   P_threshold = 1500,
                                   cooling_considered = False,
                                   T_min_cooling = None,
                                   T_max_cooling = None,
                                   heating_season_begin = None,
                                   heating_season_end = None)
        return config

    def build( self, config ): 
        self.name = config.name
        self.source_weight = config.source_weight   
        self.T_min_heating = config.T_min_heating
        self.T_max_heating = config.T_max_heating
        self.T_tolerance = config.T_tolerance
        self.P_threshold = config.P_threshold
        self.cooling_considered = config.cooling_considered
        if self.cooling_considered:
            self.T_min_cooling = config.T_min_cooling
            self.T_max_cooling = config.T_max_cooling
            self.heating_season_begin = config.heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            self.heating_season_end = config.heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
        self.state = L2HeatSmartControllerState()
        self.previous_state = L2HeatSmartControllerState()
        
    def control_cooling( self, T_control: float, T_min_cooling: float, T_max_cooling: float, l3state: Any) -> None:
        if T_control > T_max_cooling:
            #start cooling if temperature exceeds upper limit
            self.state.activate( )
            self.previous_state.activate( )

        elif T_control < T_min_cooling:
            #stop cooling if temperature goes below lower limit
            self.state.deactivate( )
            self.previous_state.deactivate( )

        else:
            if self.state.compulsory == 1:
                #use previous state if it is compulsory
                pass
            elif self.ElectricityTargetC.source_output is not None:
                #use recommendation from l3 if available and not compulsory
                self.state.state = l3state
            else:
                #use previous state if l3 was not available
                self.state = self.previous_state.clone( )
                
    def control_heating( self, T_control: float, T_min_heating: float, T_max_heating: float, l3state: Any) -> int:
        if l3state > 0:
            T_min_heating = T_min_heating + 5
        if T_control < T_min_heating:
            return 1
        return 0


    def i_save_state(self):
        self.previous_state = self.state.clone( )

    def i_restore_state(self):
        self.state = self.previous_state.clone( )

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues,  force_convergence: bool) -> None:
        if force_convergence:
            return
        # check demand, and change state of self.has_heating_demand, and self._has_cooling_demand
        T_control = stsv.get_input_value( self.ReferenceTemperatureC )  
        # if self.my_simulation_parameters.system_config.predictive == True:
        #     RunTimeSignal = stsv.get_input_value( self.l1_RunTimeSignalC )
        # else:
        #     RunTimeSignal = 0

        #get l3 recommendation if available
        electricity_target = stsv.get_input_value( self.ElectricityTargetC )
        if electricity_target >= self.P_threshold:
            l3state = 1
        else:
            l3state = 0

        #reset temperature limits if recommended from l3
        # if self.cooling_considered:
        #     if l3state == 1 :
        #         if RunTimeSignal > 0:
        #             T_min_cooling = self.T_min_cooling - self.T_tolerance
        #         else:
        #             T_min_cooling = ( self.T_min_cooling + self.T_max_cooling ) / 2
        #         T_max_cooling = self.T_max_cooling
        #     elif l3state == 0:
        #         T_max_cooling = self.T_max_cooling + self.T_tolerance
        #         T_min_cooling = self.T_min_cooling

        # if l3state == 1:
        #     # if RunTimeSignal > 0:
        #     #     T_max_heating = self.T_max_heating + self.T_tolerance
        #     # else:
        #     T_max_heating = ( self.T_min_heating + self.T_max_heating ) / 2
        #     T_min_heating = self.T_min_heating
        #     self.state.is_compulsory( )
        #     self.previous_state.is_compulsory( )
        # elif l3state == 0:
        #      T_max_heating = self.T_max_heating
        #      T_min_heating = self.T_min_heating - self.T_tolerance
        #      self.state.is_compulsory( )
        #      self.previous_state.is_compulsory( )


        # if self.cooling_considered:
        #     #check out during cooling season
        #     if timestep < self.heating_season_begin and timestep > self.heating_season_end:
        #         self.control_cooling( T_control = T_control, T_min_cooling = T_min_cooling, T_max_cooling = T_max_cooling, l3state = l3state )
        #     #check out during heating season
        #     else:
        #         self.control_heating( T_control = T_control, T_min_heating = T_min_heating, T_max_heating = T_max_heating, l3state = l3state )
        #
        # #check out during heating season
        # else:
        control_signal = self.control_heating( T_control = T_control, T_min_heating = self.T_min_heating, T_max_heating = self.T_max_heating, l3state = l3state )
        stsv.set_output_value( self.l2_DeviceSignalC, control_signal )
        
    def write_to_report( self ):
        lines = []
        lines.append("Name: {}".format(self.name + str( self.source_weight ) ) )
        lines.append("upper set temperature: {:4.0f} °C".format( ( self.T_max_heating ) ) )
        lines.append( "lower set temperature: {:4.0f} °C".format( self.T_min_heating ) )
        lines.append( "tolerance: {:4.0f} °C".format( self.T_tolerance))
        return lines




# Generic/Built-in

# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim.components.generic_heat_pump  import HeatPumpController
from hisim.components.generic_ev_charger  import EVChargerController
from typing import List, Any, Dict
from hisim import loadtypes as lt
from abc import ABC, abstractmethod
from hisim.simulationparameters import SimulationParameters


class SmartController(Component):

    def __init__(self,my_simulation_parameters: SimulationParameters, controllers: Dict[Any,Any] = {"HeatPump":["mode"], "EVCharger":["mode"]})  -> None:
        super().__init__(name="SmartController", my_simulation_parameters=my_simulation_parameters)
        self.WrappedControllers:List[Any] = []
        self.build(controllers)

    def build(self, controllers:Dict[Any,Any]) -> None:
        for controller_name in controllers:
            if "HeatPump" in controller_name:
                self.WrappedControllers.append(HeatPumpController(my_simulation_parameters=self.my_simulation_parameters))
            elif "EVCharger" in controller_name:
                self.WrappedControllers.append(EVChargerController(my_simulation_parameters=self.my_simulation_parameters))
        self.add_io()

    def connect_similar_inputs(self, components:List[Any]) -> None:
        if len(self.inputs) == 0:
            raise Exception("The component " + self.component_name + " has no inputs.")

        if isinstance(components, list) is False:
            components = [components]

        for component in components:
            if isinstance(component, Component) is False:
                raise Exception("Input variable is not a component")
            has_not_been_connected = True
            for index, controller in enumerate(self.WrappedControllers):
                for input in self.WrappedControllers[index].inputs:
                        for output in component.outputs:
                            if input.field_name == output.field_name:
                                has_not_been_connected = False
                                self.WrappedControllers[index].connect_input(self.WrappedControllers[index].field_name, component.component_name, output.field_name)
            if has_not_been_connected:
                raise Exception("No similar inputs from {} are compatible with the outputs of {}!".format(self.WrappedControllers[index].component_name, component.component_name))

    def add_io(self) -> None:
        for controller in self.WrappedControllers:
            for input in controller.inputs:
                self.inputs.append(input)
            for output in controller.outputs:
                self.outputs.append(output)

    def i_save_state(self) -> None:
        for index, controller in enumerate(self.WrappedControllers):
            self.WrappedControllers[index].i_save_state()

    def i_restore_state(self) -> None:
        for index, controller in enumerate(self.WrappedControllers):
            self.WrappedControllers[index].i_restore_state()

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        pass

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        for index, controller in enumerate(self.WrappedControllers):
            self.WrappedControllers[index].i_simulate(timestep=timestep, stsv=stsv, force_convergence=force_convergence)

    def connect_electricity(self, component: Any) -> None:
        for index, controller in enumerate(self.WrappedControllers):
            if hasattr(self.WrappedControllers[index], "ElectricityInput"):
                if isinstance(component, Component) is False:
                    raise Exception("Input has to be a component!")
                elif hasattr(component, "ElectricityOutput") is False:
                    raise Exception("Input Component does not have Electricity Output!")
                self.connect_input(self.WrappedControllers[index].ELECTRICITY_INPUT, component.component_name,
                                   component.ElectricityOutput)




""" Module for visualizing the entire system as a flow chart. """
# clean

import os
import pydot

from hisim.postprocessing.postprocessing_datatransfer import PostProcessingDataTransfer


class SystemChart:

    """ Class for generating charts that show all the components. """

    def __init__(self, ppdt: PostProcessingDataTransfer) -> None:
        """ Initizalizes the class. """
        self.ppdt: PostProcessingDataTransfer = ppdt

    def make_chart(self) -> None:
        """ Makes different charts. Entry point for the class. """
        self.make_graphviz_chart(with_labels=False, with_class_names=True, filename="System_no_Edge_with_class_labels.png")
        self.make_graphviz_chart(with_labels=False, with_class_names=False, filename="System_no_Edge_labels.png")
        self.make_graphviz_chart(with_labels=True, with_class_names=False, filename="System_with_Edge_labels.png")

    def make_graphviz_chart(self, with_labels: bool, with_class_names: bool, filename: str) -> None:
        """ Visualizes the entire system with graphviz. """
        graph = pydot.Dot(graph_type='digraph')
        graph.set_node_defaults(color='lightgray', style='filled', shape='box', fontname='Arial', fontsize='10')
        node_dict = {}
        for component in self.ppdt.wrapped_components:
            node_name = component.my_component.component_name
            if with_class_names:
                node_name = node_name + "\n" + component.my_component.__class__.__name__
            my_node = pydot.Node(node_name)
            node_dict[component.my_component.component_name] = my_node
            graph.add_node(my_node)
        edge_labels = {}
        for component in self.ppdt.wrapped_components:
            for component_input in component.component_inputs:
                if component_input.src_object_name is None:
                    continue
                node_a = node_dict[component_input.src_object_name]
                node_b = node_dict[component.my_component.component_name]
                key = (node_a, node_b)
                this_edge_label = str(component_input.src_field_name) + " -> " + component_input.field_name + " in " + component_input.unit
                this_edge_label = this_edge_label.replace("°C", "&#8451;")
                if key not in edge_labels:
                    edge_labels[key] = this_edge_label
                else:
                    edge_labels[key] = edge_labels[key] + "\\n" + this_edge_label

        for node_key, label in edge_labels.items():
            if with_labels:
                graph.add_edge(pydot.Edge(node_key[0], node_key[1], label=label))
            else:
                graph.add_edge(pydot.Edge(node_key[0], node_key[1]))
        fullpath = os.path.join(self.ppdt.simulation_parameters.result_directory, filename)
        graph.write_png(fullpath)  # noqa: no-member
